#!/usr/bin/python3
########################################################################################
# hpc_clover.py - Entry point for running CLOVER on Imperial College's HPC.            #
#                                                                                      #
# Authors: Phil Sandwell, Ben Winchester                                               #
# Copyright: Phil Sandwell, 2022                                                       #
# Date created: 29/03/2022                                                             #
# License: Open source                                                                 #
#                                                                                      #
# For more information, please email:                                                  #
#   benedict.winchester@gmail.com                                                      #
#   philip.sandwell@gmail.com                                                          #
########################################################################################
"""
hpc_clover.py - The entry point for running CLOVER on the HPC.

Imperial College London owns a series of high-performance computers. This module
provides an entry point for running CLOVER across the various HPC computers.

"""

import math
import os
import subprocess
import sys
import tempfile

from logging import Logger
from typing import Union

from ..__main__ import __version__
from ..__utils__ import (
    DONE,
    FAILED,
    LOCATIONS_FOLDER_NAME,
    BColours,
    ProgrammerJudgementFault,
    get_logger,
)
from ..fileparser import INPUTS_DIRECTORY, LOAD_INPUTS_DIRECTORY, parse_scenario_inputs
from .hpc_utils import (
    HpcOptimisation,
    HpcRunType,
    HpcSimulation,
    InvalidRunError,
    parse_args_and_hpc_input_file,
)

__all__ = ("main",)


# Clover hpc header string:
#   The ascii text to display when starting CLOVER on the HPC.
CLOVER_HPC_HEADER_STRING = """

        (((((*    /(((
        ((((((( ((((((((
   (((((((((((( ((((((((((((
   ((((((((((((*(((((((((((((       _____ _      ______      ________ _____
     *((((((((( ((((((((((((       / ____| |    / __ \\ \\    / /  ____|  __ \\
   (((((((. /((((((((((/          | |    | |   | |  | \\ \\  / /| |__  | |__) |
 ((((((((((((((((((((((((((,      | |    | |   | |  | |\\ \\/ / |  __| |  _  /
 (((((((((((*  (((((((((((((      | |____| |___| |__| | \\  /  | |____| | \\ \\
   ,(((((((. (  (((((((((((/       \\_____|______\\____/   \\/   |______|_|  \\_\\
   .((((((   (   ((((((((
             /     (((((
             ,
              ,
               (
                 (
                   (

                ___                     _      _   _  _ ___  ___
               |_ _|_ __  _ __  ___ _ _(_)__ _| | | || | _ \\/ __|
                | || '  \\| '_ \\/ -_) '_| / _` | | | __ |  _/ (__
               |___|_|_|_| .__/\\___|_| |_\\__,_|_| |_||_|_|  \\___|
                         |_|

       Continuous Lifetime Optimisation of Variable Electricity Resources
                         Copyright Phil Sandwell, 2018
{version_line}

   This version of CLOVER has been adapted for Imperial College London's HPC
  See the user guide for more information on how to use this version of CLOVER

                         For more information, contact
                   Phil Sandwell (philip.sandwell@gmail.com),
                    Hamish Beath (hamishbeath@outlook.com),
               or Ben Winchester (benedict.winchester@gmail.com)

"""

# Hpc submission script filename:
#   The name of the HPC script submission file.
HPC_SUBMISSION_SCRIPT_FILENAME: str = "hpc.sh"

# Hpc submission script filepath:
#   The path to the HPC script submission file.
HPC_SUBMISSION_SCRIPT_FILEPATH: str = os.path.join(
    "bin", HPC_SUBMISSION_SCRIPT_FILENAME
)

# Logger name:
#   The name to use for the logger for this script.
LOGGER_NAME: str = "hpc_clover"


def _check_run(logger: Logger, hpc_run: Union[HpcOptimisation, HpcSimulation]) -> bool:
    """
    Checks that the HPC run is valid.

    Inputs:
        - logger:
            The logger to use for the run.
        - hpc_run:
            The HPC run to carry out.

    Outputs:
        - Whether the run is valid or not.

    """

    # Check that the locations folder exists.
    if not os.path.isdir(os.path.join(LOCATIONS_FOLDER_NAME, hpc_run.location)):
        logger.error(
            "%sLocation '%s' does not exist.%s",
            BColours.fail,
            hpc_run.location,
            BColours.endc,
        )
        return False

    # Check that the total load file exists if specified.
    if hpc_run.total_load_file is not None and not os.path.isfile(
        os.path.join(
            LOCATIONS_FOLDER_NAME,
            hpc_run.location,
            INPUTS_DIRECTORY,
            LOAD_INPUTS_DIRECTORY,
            hpc_run.total_load_file,
        )
    ):
        logger.error(
            "%sThe total run file '%s' could not be found in the load inputs directory."
            "%s",
            BColours.fail,
            hpc_run.total_load_file,
            BColours.endc,
        )
        return False

    # Check that the scenario exists as a scenario.
    if hpc_run.type == HpcRunType.SIMULATION:
        if not isinstance(hpc_run, HpcSimulation):
            logger.error(
                "%sRun marked as simulation but was processed as an optimisation.%s",
                BColours.fail,
                BColours.endc,
            )
            raise ProgrammerJudgementFault(
                "hpc_clover.py", "Failure processing run as a simulation."
            )

        # Parse the scenario files for the location.
        logger.info("%sParsing scenario input file.%s", BColours.fail, BColours.endc)
        _, _, scenarios, _ = parse_scenario_inputs(
            os.path.join(LOCATIONS_FOLDER_NAME, hpc_run.location, INPUTS_DIRECTORY),
            logger,
        )

        if hpc_run.scenario not in {scenario.name for scenario in scenarios}:
            logger.error(
                "%sScenario '%s' not in the scenarios file.%s",
                BColours.fail,
                hpc_run.scenario,
                BColours.endc,
            )
            return False

    # Returns false if not.
    return True


def main(args) -> None:
    """
    Wrapper around CLOVER when run on the HPC.

    """

    version_string = f"Version {__version__}"
    print(
        CLOVER_HPC_HEADER_STRING.format(
            version_line=(
                " " * (40 - math.ceil(len(version_string) / 2))
                + version_string
                + " " * (40 - math.floor(len(version_string) / 2))
            )
        )
    )

    logger = get_logger(LOGGER_NAME, False)
    logger.info("HPC-CLOVER script called.")
    logger.info("Arguments: %s", ", ".join(args))

    # Call the utility module to parse the HPC run information.
    run_file, runs, verbose, walltime = parse_args_and_hpc_input_file(args, logger)
    logger.info("Command-line arguments successfully parsed. Run file: %s", run_file)
    logger.debug("Runs:\n%s- ", "\n- ".join(str(run) for run in runs))

    # Check that all of the runs are valid.
    print("Checking HPC runs .............................................    ", end="")
    logger.info("Checking all run files are valid.")
    if not all(_check_run(logger, run) for run in runs):
        logger.error(
            "%sNot all HPC runs were valid, exiting.%s", BColours.fail, BColours.endc
        )
        print(FAILED)
        raise InvalidRunError("Not all HPC runs were valid, see logs for details.")

    print(DONE)
    logger.info("All HPC runs valid.")

    # Parse the default HPC job submission script.
    print("Processing HPC job submission script ..........................    ", end="")
    logger.info("Parsing base HPC job submission script.")
    try:
        with open(HPC_SUBMISSION_SCRIPT_FILEPATH, "r") as f:
            hpc_submission_script_file_contents = f.read()
    except FileNotFoundError:
        logger.error(
            "%sHPC job submission file not found. Check that the file, '%s', has not "
            "been removed.%s",
            BColours.fail,
            HPC_SUBMISSION_SCRIPT_FILEPATH,
            BColours.endc,
        )
        print(FAILED)
        raise

    print(DONE)
    logger.info("HPC job submission file successfully parsed.")

    hpc_submission_script_file_contents = hpc_submission_script_file_contents.format(
        NUM_RUNS=len(runs),
        RUNS_FILE=run_file,
        VERBOSE=("--verbose" if verbose else ""),
        WALLTIME=walltime,
    )
    logger.info(
        "HPC job submission script updated with %s runs, %s walltime.",
        len(runs),
        walltime,
    )

    # Setup the HPC job submission script.
    with tempfile.TemporaryDirectory() as tmpdirname:
        hpc_submission_script_filepath = os.path.join(
            tmpdirname, HPC_SUBMISSION_SCRIPT_FILENAME
        )

        # Write the submission script file.
        logger.info("Writing temporary HPC submission script.")
        with open(hpc_submission_script_filepath, "w") as f:
            f.write(hpc_submission_script_file_contents)

        logger.info("HPC job submission script successfully submitted.")

        # Update permissions on the file.
        os.chmod(hpc_submission_script_filepath, 0o775)
        logger.info("HPC job submission script permissions successfully updated.")

        # Submit the script to the HPC.
        logger.info("Submitting CLOVER jobs to the HPC.")
        print("Sending jobs to the HPC:")
        try:
            subprocess.run(["qsub", hpc_submission_script_filepath], check=False)
        except Exception:  # pylint: disable=broad-except
            logger.error("Failed. See logs for details.")
            print(
                "Sending jobs to the HPC .......................................    "
                f"{FAILED}"
            )
            raise
        print(
            f"Sending jobs to the HPC .......................................    {DONE}"
        )
        logger.info("HPC runs submitted. Exiting.")


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/python3
########################################################################################
# __main__.py - Main module for CLOVER.                                                #
#                                                                                      #
# Author: Ben Winchester                                                               #
# Copyright: Ben Winchester, 2021                                                      #
# Date created: 13/07/2021                                                             #
# License: Open source                                                                 #
########################################################################################
"""
__main__.py - The main module for CLOVER.

CLOVER (Continuous Lifetime Optimisation of Variable Electricity Resources) can evaluate
and optimise minigrid systems, determining whether a demand is met whilst minimising
environmental and economic impacts. The main flow of CLOVER can be executed by running
the clover module from the command-line interface.

"""

__version__ = "5.0.8"

import datetime
import logging
import math
import os
import sys

from argparse import Namespace
from typing import Any, Dict, List, Optional, Set, Tuple

import pandas as pd  # pylint: disable=import-error

from tqdm import tqdm

from . import analysis, argparser
from .conversion.conversion import WaterSource
from .fileparser import (
    INPUTS_DIRECTORY,
    KEROSENE_TIMES_FILE,
    KEROSENE_USAGE_FILE,
    parse_input_files,
)
from .generation import solar, weather, wind
from .load import load
from .mains_supply import grid, water_source
from .scripts import new_location
from .simulation import energy_system

from .optimisation.__utils__ import save_optimisation
from .optimisation.appraisal import appraise_system
from .optimisation.optimisation import multiple_optimisation_step
from .printer import generate_optimisation_string, generate_simulation_string

from .__utils__ import (
    BColours,
    DONE,
    FAILED,
    InternalError,
    Location,
    ResourceType,
    SystemAppraisal,
    get_logger,
    InputFileError,
    LOCATIONS_FOLDER_NAME,
    LOGGER_DIRECTORY,
    OperatingMode,
    save_simulation,
)
from .simulation.__utils__ import check_scenario

__all__ = ("main",)

# Auto-generated-files directory:
#   The name of the directory in which to save auto-generated files, relative to the
# root of the location.
AUTO_GENERATED_FILES_DIRECTORY = "auto_generated"

# Clover header string:
#   The ascii text to display when starting CLOVER.
CLOVER_HEADER_STRING = """
\033[38;5;40m
        (((((*    /(((
        ((((((( ((((((((
   (((((((((((( ((((((((((((
   ((((((((((((*(((((((((((((       _____ _      ______      ________ _____
     *((((((((( ((((((((((((       / ____| |    / __ \\ \\    / /  ____|  __ \\
   (((((((. /((((((((((/          | |    | |   | |  | \\ \\  / /| |__  | |__) |
 ((((((((((((((((((((((((((,      | |    | |   | |  | |\\ \\/ / |  __| |  _  /
 (((((((((((*  (((((((((((((      | |____| |___| |__| | \\  /  | |____| | \\ \\
   ,(((((((. (  (((((((((((/       \\_____|______\\____/   \\/   |______|_|  \\_\\
   .((((((   (   ((((((((
             /     (((((
             ,
              ,
               (
                 (
                   (
\033[0m


       Continuous Lifetime Optimisation of Variable Electricity Resources
                         Copyright Phil Sandwell, 2018
{version_line}

                         For more information, contact
                   Phil Sandwell (philip.sandwell@gmail.com),
                    Hamish Beath (hamishbeath@outlook.com),
               or Ben Winchester (benedict.winchester@gmail.com)

"""

# Debug string:
#   Text to display when debug mode is selected.
DEBUG_STRING: str = """{okblue}                        CLOVER is running in debug mode.
          For more information on debug mode, consult the user guide.
    If you did not intend to use debug mode, re-run without the debug flag.
{endc}"""

# Logger name:
#   The name to use for the main logger for CLOVER
LOGGER_NAME = "clover"

# Number of Workers:
#   The number of CPUs to use, which dictates the number of workers to use for parllel
#   jobs.
NUM_WORKERS = 8

# Outputs folder:
#   The folder into which outputs should be saved.
OUTPUTS_FOLDER = "outputs"

# Optimisation outputs folder:
#   The output folder into which to save the optimisation outputs.
OPTIMISATION_OUTPUTS_FOLDER = os.path.join(OUTPUTS_FOLDER, "optimisation_outputs")

# Simulation outputs folder:
#   The folder into which outputs should be saved.
SIMULATION_OUTPUTS_FOLDER = os.path.join(OUTPUTS_FOLDER, "simulation_outputs")


def _get_operating_mode(parsed_args: Namespace) -> OperatingMode:
    """
    Determine the operating mode for CLOVER based on the command-line arguments.

    Inputs:
        - parsed_args:
            The parsed command-line arguments.

    Outputs:
        - The operating mode to use for the run.

    """

    # Try to determine the operating mode.
    if parsed_args.simulation:
        return OperatingMode.SIMULATION
    if parsed_args.optimisation:
        return OperatingMode.OPTIMISATION
    return OperatingMode.PROFILE_GENERATION


def _prepare_location(location: str, logger: logging.Logger) -> None:
    """
    Prepares the location and raises an error if the location cannot be found.

    Inputs:
        - location
            The name of the location to check.

    Raises:
        - FileNotFoundError:
            Raised if the location cannot be found.

    """

    if not os.path.isdir(os.path.join(LOCATIONS_FOLDER_NAME, location)):
        logger.error(
            "%sThe specified location, '%s', does not exist. Try running the "
            "'new_location' script to ensure all necessary files and folders are "
            "present.%s",
            BColours.fail,
            location,
            BColours.endc,
        )
        raise FileNotFoundError(f"The location, {location}, could not be found.")

    if not os.path.isfile(
        os.path.join(
            LOCATIONS_FOLDER_NAME, location, INPUTS_DIRECTORY, KEROSENE_TIMES_FILE
        )
    ):
        logger.info(
            "%sThe specified location, '%s', does not contain a kerosene times file. "
            "The auto-generation script will be run to replace the lost file.%s",
            BColours.warning,
            location,
            BColours.endc,
        )
        new_location.create_new_location(None, location, logger, True)
        logger.info("%s succesfully updated with missing files.", location)


def _prepare_water_system(
    available_conventional_sources: Set[str],
    auto_generated_files_directory: str,
    device_utilisations: Dict[load.Device, pd.DataFrame],
    disable_tqdm: bool,
    location: Location,
    logger: logging.Logger,
    parsed_args: Namespace,
    resource_type: ResourceType,
    water_source_times: Dict[WaterSource, pd.DataFrame],
) -> Tuple[
    Dict[WaterSource, pd.DataFrame], Dict[str, pd.DataFrame], pd.DataFrame, pd.DataFrame
]:
    """
    Prepares the conventional-water system.

    Inputs:
        - available_conventional_sources:
            The `list` of available conventional sources for the system.
        - auto_generated_files_directory:
            The directory into which auto-generated files should be saved.
        - device_utilisations:
            The utilisation profile for each :class:`load.Device` being modelled.
        - location:
            The :class:`Location` being considered.
        - logger:
            The :class:`logging.Logger` to use for the run.
        - parsed_args:
            The parsed command-line arguments.
        - resource_type:
            The :class:`ResourceType` being considered.
        - water_source_times:
            The availability profile of each :class:`WaterSource` being considered.

    Outputs:
        - conventional_water_source_profiles:
            The availability profiles of the conventional water sources being
            considered.
        - disable_tqdm:
            Whether to disable the tqdm progress bars (True) or display them (False).
        - initial_loads:
            The initial hourly loads placed on the conventional water system.
        - total_load:
            The total water load for each load type placed on the conventional water
            system.
        - yearly_load_statistics:
            The yearly load statistics for the conventional water system being
            considered.

    """

    # Raise an error if there are no water devices specified.
    if (
        resource_type == ResourceType.HOT_CLEAN_WATER
        and (
            len(
                {
                    device
                    for device in device_utilisations
                    if device.hot_water_usage is not None
                }
            )
            == 0
        )
        or resource_type == ResourceType.CLEAN_WATER
        and (
            len(
                {
                    device
                    for device in device_utilisations
                    if device.clean_water_usage is not None
                }
            )
            == 0
        )
    ):
        raise InputFileError(
            "devices input flie",
            f"No {resource_type.value} input devices were specified despite the "
            + "scenario containing a clean-water system.",
        )

    try:
        (
            initial_loads,
            total_load,
            yearly_load_statistics,
        ) = load.process_load_profiles(
            auto_generated_files_directory,
            device_utilisations,
            disable_tqdm,
            location,
            logger,
            parsed_args.regenerate,
            resource_type,
        )
    except InputFileError:
        print(
            "Generating necessary profiles .................................    "
            + f"{FAILED}"
        )
        raise
    except Exception as e:
        print(
            "Generating necessary profiles .................................    "
            + f"{FAILED}"
        )
        logger.error(
            "%sAn unexpected error occurred generating the %s load profiles. See %s "
            "for details: %s%s",
            BColours.fail,
            resource_type.value,
            f"{os.path.join(LOGGER_DIRECTORY, LOGGER_NAME)}.log",
            str(e),
            BColours.endc,
        )
        raise

    # Generate the conventional-clean-water source availability profiles.
    logger.info(
        "Generating conventional %s water-source availability profiles.",
        resource_type.value,
    )
    try:
        conventional_water_source_profiles = (
            water_source.get_lifetime_water_source_status(
                disable_tqdm,
                os.path.join(auto_generated_files_directory, resource_type.value),
                resource_type.value.split("_")[0],
                location,
                logger,
                parsed_args.regenerate,
                water_source_times,
            )
        )
    except InputFileError:
        print(
            "Generating necessary profiles .................................    "
            + f"{FAILED}"
        )
        raise
    except Exception as e:
        print(
            "Generating necessary profiles .................................    "
            + f"{FAILED}"
        )
        logger.error(
            "%sAn unexpected error occurred generating the conventional %s "
            "water-source profiles. See %s for details: %s%s",
            BColours.fail,
            resource_type.value,
            f"{os.path.join(LOGGER_DIRECTORY, LOGGER_NAME)}.log",
            str(e),
            BColours.endc,
        )
        raise

    logger.info(
        "Conventional %s water sources successfully parsed.", resource_type.value
    )
    logger.debug(
        "Conventional %s water sources: %s",
        resource_type.value,
        ", ".join([str(source) for source in conventional_water_source_profiles]),
    )

    conventional_water_source_profiles = {
        key: value
        for key, value in conventional_water_source_profiles.items()
        if key in available_conventional_sources
    }

    return (
        conventional_water_source_profiles,
        initial_loads,
        total_load,
        yearly_load_statistics,
    )


def main(  # pylint: disable=too-many-locals, too-many-statements
    args: List[Any], disable_tqdm: bool = False, run_number: Optional[int] = None
) -> None:
    """
    The main module for CLOVER executing all functionality as appropriate.

    Inputs:
        - args
            The command-line arguments, passed in as a list.
        - disable_tqdm:
            Whether to disable the tqdm progress bars (True) or display them (False).
        - run_number:
            Used to differentiate between runs if multiple runs are being carried out
            for the same location.

    """

    # Parse the command-line arguments and instantiate the logger.
    parsed_args = argparser.parse_args(args)
    run_number_string: str = f"_{run_number}" if run_number is not None else ""
    logger = get_logger(
        f"{parsed_args.location}_{LOGGER_NAME}{run_number_string}",
        parsed_args.verbose,
    )
    logger.info("CLOVER run initiated. Options specified: %s", " ".join(args))

    # Validate the command-line arguments.
    logger.info("Command-line arguments successfully parsed.")

    if not argparser.validate_args(logger, parsed_args):
        logger.error(
            "%sInvalid command-line arguments. Check that all required arguments have "
            "been specified correctly. See %s for details.%s",
            BColours.fail,
            f"{os.path.join(LOGGER_DIRECTORY, LOGGER_NAME)}.log",
            BColours.endc,
        )
        raise ValueError(
            "The command-line arguments were invalid. See "
            f"{os.path.join(LOGGER_DIRECTORY, LOGGER_NAME)}.log for details."
        )

    logger.info("Command-line arguments successfully validated.")

    version_string = f"Version {__version__}"
    print(
        CLOVER_HEADER_STRING.format(
            version_line=(
                " " * (40 - math.ceil(len(version_string) / 2))
                + version_string
                + " " * (40 - math.floor(len(version_string) / 2))
            )
        )
    )

    if parsed_args.debug:
        print(DEBUG_STRING.format(okblue=BColours.okblue, endc=BColours.endc))

    # Define common variables.
    auto_generated_files_directory = os.path.join(
        LOCATIONS_FOLDER_NAME,
        parsed_args.location,
        AUTO_GENERATED_FILES_DIRECTORY,
    )

    # If the output filename is not provided, then generate it.
    simulation_output_directory = os.path.join(
        LOCATIONS_FOLDER_NAME,
        parsed_args.location,
        SIMULATION_OUTPUTS_FOLDER,
    )
    optimisation_output_directory = os.path.join(
        LOCATIONS_FOLDER_NAME, parsed_args.location, OPTIMISATION_OUTPUTS_FOLDER
    )

    # Determine the operating mode for the run.
    operating_mode = _get_operating_mode(parsed_args)
    if operating_mode == OperatingMode.SIMULATION:
        output_directory: Optional[str] = simulation_output_directory
        logger.info(
            "A single CLOVER simulation will be run for locatation '%s'",
            parsed_args.location,
        )
        print(
            f"A single CLOVER simulation will be run for {parsed_args.location}"
            + (
                f" {BColours.okblue}in debug mode{BColours.endc}"
                if parsed_args.debug
                else ""
            )
        )
    if operating_mode == OperatingMode.OPTIMISATION:
        output_directory = optimisation_output_directory
        logger.info(
            "A CLOVER optimisation will be run for location '%s'", parsed_args.location
        )
        print(
            f"A CLOVER optimisation will be run for {parsed_args.location}"
            + (
                f" {BColours.okblue}in debug mode{BColours.endc}"
                if parsed_args.debug
                else ""
            )
        )
    if operating_mode == OperatingMode.PROFILE_GENERATION:
        output_directory = None
        logger.info("No CLI mode was specified, CLOVER will only generate profiles.")
        print(
            "Neither `simulation` or `optimisation` specified, running profile "
            f"generation only for {parsed_args.location}"
            + (
                f" {BColours.okblue}in debug mode{BColours.endc}"
                if parsed_args.debug
                else ""
            )
        )

    # If the output folder already exists, then confirm from the user that they wish to
    # overwrite its contents.
    if parsed_args.output is not None:
        if output_directory is None:
            logger.error(
                "%sCannot specify an output directory if only profile generation is "
                "taking place.%s",
                BColours.fail,
                BColours.endc,
            )
            raise Exception(
                "The `output` flag can only be used if a simulation or optimisation "
                "output is expected."
            )
        if os.path.isdir(os.path.join(output_directory, parsed_args.output)):
            try:
                confirm_overwrite = {"y": True, "n": False, "Y": True, "N": False}[
                    input(
                        f"Output folder, {parsed_args.output}, already exists, Overwrite? [y/n] "
                    )
                ]
            except KeyError:
                logger.error(
                    "Either 'y' or 'n' must be specified to confirm overwrite. Quitting."
                )
                raise
            if confirm_overwrite:
                output: str = str(parsed_args.output)
            else:
                output = input("Specify new output folder name: ")
            print(f"\nOutput directory {output} will be used for simulation results.")
        else:
            output = str(parsed_args.output)
    else:
        output = datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S")

    # Verify the location as containing all the required files.
    print("Verifying location information ................................    ", end="")
    logger.info("Checking location %s.", parsed_args.location)
    try:
        _prepare_location(parsed_args.location, logger)
    except FileNotFoundError:
        print(FAILED)
        logger.error(
            "%sThe location, '%s', is missing files. Try running the `new_location` "
            "script to identify missing files. See %s for details.%s",
            BColours.fail,
            parsed_args.location,
            f"{os.path.join(LOGGER_DIRECTORY, LOGGER_NAME)}.log",
            BColours.endc,
        )
        raise

    # Parse the various input files.
    print(
        f"{DONE}\nParsing input files ...........................................    ",
        end="",
    )
    logger.info("Parsing input files.")

    try:
        (
            converters,
            device_utilisations,
            minigrid,
            finance_inputs,
            generation_inputs,
            ghg_inputs,
            grid_times,
            location,
            optimisation_inputs,
            optimisations,
            scenarios,
            simulations,
            electric_load_profile,
            water_source_times,
            input_file_info,
        ) = parse_input_files(
            parsed_args.debug,
            parsed_args.electric_load_profile,
            parsed_args.location,
            logger,
            parsed_args.optimisation_inputs_file,
        )
    except FileNotFoundError as e:
        print(FAILED)
        logger.error(
            "%sNot all input files present. See %s for details: %s%s",
            BColours.fail,
            f"{os.path.join(LOGGER_DIRECTORY, LOGGER_NAME)}.log",
            str(e),
            BColours.endc,
        )
        raise
    except InputFileError as e:
        print(FAILED)
        logger.error("Input file error occured: %s", str(e))
        raise
    except Exception as e:
        print(FAILED)
        logger.error(
            "%sAn unexpected error occured parsing input files. See %s for details: "
            "%s%s",
            BColours.fail,
            f"{os.path.join(LOGGER_DIRECTORY, LOGGER_NAME)}.log",
            str(e),
            BColours.endc,
        )
        raise

    logger.info("All input files successfully parsed.")
    print(DONE)

    print("Generating necessary profiles", end="\n")

    # Determine the number of background tasks to carry out.
    num_ninjas: int = (
        1
        + (1 if any(scenario.pv_t for scenario in scenarios) else 0)
        + (
            1
            if any(scenario.desalination_scenario for scenario in scenarios) is not None
            else 0
        )
    )

    # Generate and save the wind data for each year as a background task.
    if any(scenario.pv_t for scenario in scenarios):
        logger.info("Beginning wind-data fetching.")
        wind_data_thread: Optional[wind.WindDataThread] = wind.WindDataThread(
            os.path.join(auto_generated_files_directory, "wind"),
            generation_inputs,
            location,
            f"{parsed_args.location}_{wind.WIND_LOGGER_NAME}",
            parsed_args.refetch,
            num_ninjas,
            parsed_args.verbose,
        )
        if wind_data_thread is None:
            raise InternalError("Wind data thread failed to successfully instantiate.")
        wind_data_thread.start()
        logger.info(
            "Wind-data thread successfully instantiated. See %s for details.",
            f"{os.path.join(LOGGER_DIRECTORY, wind.WIND_LOGGER_NAME)}.log",
        )
    else:
        wind_data_thread = None

    # Generate and save the weather data for each year as a background task.
    if any(scenario.desalination_scenario is not None for scenario in scenarios):
        # Set up the system to call renewables.ninja at a slower rate.
        logger.info("Begining weather-data fetching.")
        weather_data_thread: Optional[
            weather.WeatherDataThread
        ] = weather.WeatherDataThread(
            os.path.join(auto_generated_files_directory, "weather"),
            generation_inputs,
            location,
            f"{parsed_args.location}_{weather.WEATHER_LOGGER_NAME}",
            parsed_args.refetch,
            num_ninjas,
            parsed_args.verbose,
        )
        if weather_data_thread is None:
            raise InternalError(
                "Weather data thread failed to successfully instantiate."
            )
        weather_data_thread.start()
        logger.info(
            "Weather-data thread successfully instantiated. See %s for details.",
            f"{os.path.join(LOGGER_DIRECTORY, weather.WEATHER_LOGGER_NAME)}.log",
        )
    else:
        weather_data_thread = None

    # Generate and save the solar data for each year as a background task.
    logger.info("Beginning solar-data fetching.")
    solar_data_thread = solar.SolarDataThread(
        os.path.join(auto_generated_files_directory, "solar"),
        generation_inputs,
        location,
        f"{parsed_args.location}_{solar.SOLAR_LOGGER_NAME}{run_number_string}",
        parsed_args.refetch,
        minigrid.pv_panel,
        num_ninjas,
        parsed_args.verbose,
    )
    solar_data_thread.start()
    logger.info(
        "Solar-data thread successfully instantiated. See %s for details.",
        f"{os.path.join(LOGGER_DIRECTORY, solar.SOLAR_LOGGER_NAME)}.log",
    )

    # Generate and save the device-ownership profiles.
    logger.info("Processing device informaiton.")
    # load_logger = get_logger(load.LOAD_LOGGER_NAME)

    initial_electric_hourly_loads: Optional[Dict[str, pd.DataFrame]] = None
    total_electric_load: Optional[pd.DataFrame] = None
    electric_yearly_load_statistics: Optional[pd.DataFrame] = None

    if any(ResourceType.ELECTRIC in scenario.resource_types for scenario in scenarios):
        try:
            (
                initial_electric_hourly_loads,
                total_electric_load,
                electric_yearly_load_statistics,
            ) = load.process_load_profiles(
                auto_generated_files_directory,
                device_utilisations,
                disable_tqdm,
                location,
                logger,
                parsed_args.regenerate,
                load.ResourceType.ELECTRIC,
                electric_load_profile,
            )
        except InputFileError:
            print(
                "Generating necessary profiles .................................    "
                + f"{FAILED}"
            )
            raise
        except Exception as e:
            print(
                "Generating necessary profiles .................................    "
                + f"{FAILED}"
            )
            logger.error(
                "%sAn unexpected error occurred generating the load profiles. See %s for "
                "details: %s%s",
                BColours.fail,
                f"{os.path.join(LOGGER_DIRECTORY, LOGGER_NAME)}.log",
                str(e),
                BColours.endc,
            )
            raise

    clean_water_yearly_load_statistics: pd.DataFrame  # pylint: disable=unused-variable
    conventional_cw_source_profiles: Optional[Dict[WaterSource, pd.DataFrame]] = None
    initial_cw_hourly_loads: Optional[Dict[str, pd.DataFrame]] = None
    total_cw_load: Optional[pd.DataFrame] = None

    if any(scenario.desalination_scenario is not None for scenario in scenarios):
        # Create a set of all the conventional clean-water sources available.
        # @ BenWinchester - Repair conventional sources logic.
        conventional_sources: Set[str] = {
            source
            for scenario in scenarios
            if scenario.desalination_scenario is not None
            for source in scenario.desalination_scenario.clean_water_scenario.conventional_sources
        }

        # Generate the clean-water load profiles.
        (
            conventional_cw_source_profiles,
            initial_cw_hourly_loads,
            total_cw_load,
            clean_water_yearly_load_statistics,
        ) = _prepare_water_system(
            conventional_sources,
            auto_generated_files_directory,
            device_utilisations,
            disable_tqdm,
            location,
            logger,
            parsed_args,
            ResourceType.CLEAN_WATER,
            water_source_times,
        )

    conventional_hw_source_profiles: Dict[  # pylint: disable=unused-variable
        WaterSource, pd.DataFrame
    ]
    hot_water_yearly_load_statistics: pd.DataFrame  # pylint: disable=unused-variable
    initial_hw_hourly_loads: Optional[Dict[str, pd.DataFrame]] = None
    total_hw_load: Optional[pd.DataFrame] = None

    if any(scenario.hot_water_scenario is not None for scenario in scenarios):
        # Create a set of all the conventional hot-water sources available.
        # @ BenWinchester - Repair conventional sources logic.
        conventional_sources = {
            source
            for scenario in scenarios
            if scenario.hot_water_scenario is not None
            for source in scenario.hot_water_scenario.conventional_sources
        }

        (
            conventional_hw_source_profiles,
            initial_hw_hourly_loads,
            total_hw_load,
            hot_water_yearly_load_statistics,
        ) = _prepare_water_system(
            conventional_sources,
            auto_generated_files_directory,
            device_utilisations,
            disable_tqdm,
            location,
            logger,
            parsed_args,
            ResourceType.HOT_CLEAN_WATER,
            water_source_times,
        )

    # Assemble a means of storing the relevant loads.
    total_loads: Dict[ResourceType, Optional[pd.DataFrame]] = {
        ResourceType.CLEAN_WATER: total_cw_load,
        ResourceType.ELECTRIC: 0.001 * total_electric_load,  # type: ignore
        ResourceType.HOT_CLEAN_WATER: total_hw_load,
    }

    # Generate the grid-availability profiles if relevant.
    if any(scenario.grid for scenario in scenarios):
        logger.info("Generating grid-availability profiles.")
        try:
            grid.get_lifetime_grid_status(
                disable_tqdm,
                os.path.join(auto_generated_files_directory, "grid"),
                grid_times,
                logger,
                location.max_years,
            )
        except InputFileError:
            print(
                "Generating necessary profiles .................................    "
                + f"{FAILED}"
            )
            raise
        except Exception as e:
            print(
                "Generating necessary profiles .................................    "
                + f"{FAILED}"
            )
            logger.error(
                "%sAn unexpected error occurred generating the grid profiles. See %s for "
                "details: %s%s",
                BColours.fail,
                f"{os.path.join(LOGGER_DIRECTORY, LOGGER_NAME)}.log",
                str(e),
                BColours.endc,
            )
            raise

        logger.info("Grid-availability profiles successfully generated.")

    else:
        logger.info("Grid disabled, no grid profiles to be generated.")

    # Wait for all threads to finish before proceeding.
    logger.info("Waiting for all setup threads to finish before proceeding.")
    solar_data_thread.join()
    if weather_data_thread is not None:
        weather_data_thread.join()
    if wind_data_thread is not None:
        wind_data_thread.join()
    logger.info("All setup threads finished.")

    logger.info("Generating and saving total solar output file.")
    total_solar_data = solar.total_solar_output(
        os.path.join(auto_generated_files_directory, "solar"),
        parsed_args.regenerate,
        generation_inputs["start_year"],
        location.max_years,
    )
    logger.info("Total solar output successfully computed and saved.")

    if any(scenario.desalination_scenario is not None for scenario in scenarios) or any(
        scenario.hot_water_scenario is not None for scenario in scenarios
    ):
        logger.info("Generating and saving total weather output file.")
        total_weather_data = (  # pylint: disable=unused-variable
            weather.total_weather_output(
                os.path.join(auto_generated_files_directory, "weather"),
                parsed_args.regenerate,
                generation_inputs["start_year"],
                location.max_years,
            )
        )
        logger.info("Total weather output successfully computed and saved.")

    if any(scenario.pv_t for scenario in scenarios):
        logger.info("Generating and saving total wind data output file.")
        total_wind_data: Optional[pd.DataFrame] = wind.total_wind_output(
            os.path.join(auto_generated_files_directory, "wind"),
            parsed_args.regenerate,
            generation_inputs["start_year"],
            location.max_years,
        )
        logger.info("Total wind output successfully computed and saved.")
    else:
        total_wind_data = None

    logger.info(
        "Setup complete, continuing to CLOVER %s.",
        "main flow"
        if operating_mode == OperatingMode.PROFILE_GENERATION
        else operating_mode.value,
    )

    print(
        f"Generating necessary profiles .................................    {DONE}",
        end="\n",
    )

    # Load the relevant kerosene profile.
    try:
        with open(
            os.path.join(auto_generated_files_directory, KEROSENE_USAGE_FILE), "r"
        ) as f:
            kerosene_usage = pd.read_csv(f, header=None, index_col=False)
    except FileNotFoundError:
        logger.error(
            "%sKerosene usage file '%s' could not be found in '%s'. Check that this "
            "file has not been deleted.%s",
            BColours.fail,
            KEROSENE_USAGE_FILE,
            auto_generated_files_directory,
            BColours.endc,
        )
        raise

    # Remove the index from the file.
    kerosene_usage.reset_index(drop=True)

    # Determine whether any default sizes have been overrided.
    overrided_default_sizes: bool = (
        minigrid.pv_panel.pv_unit_overrided
        if minigrid.pv_panel is not None
        else False or minigrid.battery.storage_unit
        if minigrid.battery is not None
        else False
    )

    # Run a simulation or optimisation as appropriate.
    if operating_mode == OperatingMode.SIMULATION:
        print(
            f"Beginning CLOVER simulation runs {'.' * 30}    ",
            end="\n",
        )

        simulation_times: List[str] = []

        # Determine the scenario to use for the simulation.
        try:
            scenario = [
                scenario
                for scenario in scenarios
                if scenario.name == parsed_args.scenario
            ][0]
        except IndexError:
            logger.error(
                "%sUnable to locate scenario '%s' from scenarios.%s",
                BColours.fail,
                parsed_args.scenario,
                BColours.endc,
            )
            raise

        logger.info("Scenario '%s' successfully determined.", parsed_args.scenario)

        logger.info("Checking scenario parameters.")
        check_scenario(logger, minigrid, operating_mode, parsed_args, scenario)
        logger.info("Scenario parameters valid.")

        logger.info("Loading grid profile.")
        grid_profile = grid.load_grid_profile(
            auto_generated_files_directory, logger, scenario
        )
        logger.info("Grid '%s' profile successfully loaded.", scenario.grid_type)

        simulation_string: str = generate_simulation_string(
            minigrid, overrided_default_sizes, parsed_args, scenario
        )
        print(f"Running a simulation with:\n{simulation_string}")

        for simulation_number, simulation in enumerate(
            tqdm(
                simulations, desc="simulations", disable=disable_tqdm, unit="simulation"
            ),
            1,
        ):
            logger.info(
                "Carrying out simulation %s of %s.", simulation_number, len(simulations)
            )
            try:
                (
                    time_delta,
                    system_performance_outputs,
                    system_details,
                ) = energy_system.run_simulation(
                    parsed_args.clean_water_pvt_system_size
                    if parsed_args.clean_water_pvt_system_size is not None
                    else 0,
                    conventional_cw_source_profiles,
                    converters,
                    disable_tqdm,
                    parsed_args.storage_size,
                    grid_profile,
                    parsed_args.hot_water_pvt_system_size
                    if parsed_args.hot_water_pvt_system_size is not None
                    else 0,
                    total_solar_data[solar.SolarDataType.TOTAL_IRRADIANCE.value],
                    kerosene_usage,
                    location,
                    logger,
                    minigrid,
                    parsed_args.num_clean_water_tanks,
                    parsed_args.num_hot_water_tanks,
                    total_solar_data[solar.SolarDataType.ELECTRICITY.value]
                    * minigrid.pv_panel.pv_unit,
                    parsed_args.pv_system_size
                    if parsed_args.pv_system_size is not None
                    else 0,
                    scenario,
                    simulation,
                    total_solar_data[solar.SolarDataType.TEMPERATURE.value],
                    total_loads,
                    total_wind_data[wind.WindDataType.WIND_SPEED.value]
                    if total_wind_data is not None
                    else None,
                )
            except Exception as e:
                print(f"Beginning CLOVER simulation runs {'.' * 30}    {FAILED}")
                logger.error(
                    "%sAn unexpected error occurred running a CLOVER simulation. See "
                    "%s for "
                    "details: %s%s",
                    BColours.fail,
                    f"{os.path.join(LOGGER_DIRECTORY, LOGGER_NAME)}.log",
                    str(e),
                    BColours.endc,
                )
                raise

            # Add the time to the counter.
            simulation_times.append(
                "{0:.3f} s/year".format(  # pylint: disable=consider-using-f-string
                    (time_delta.seconds + time_delta.microseconds * 0.000001)
                    / (simulation.end_year - simulation.start_year)
                )
            )

            # Add the input file information to the system details file.
            system_details.file_information = input_file_info

            # Compute the key results.
            key_results = analysis.get_key_results(  # type: ignore
                grid_profile,
                simulation.end_year - simulation.start_year,
                system_performance_outputs,
                total_solar_data[solar.SolarDataType.ELECTRICITY.value]
                * minigrid.pv_panel.pv_unit
                * scenario.pv,
            )

            if parsed_args.analyse:
                if not parsed_args.skip_plots:
                    # Generate and save the various plots.
                    analysis.plot_outputs(  # type: ignore
                        grid_times[scenario.grid_type],
                        grid_profile,
                        initial_cw_hourly_loads,
                        initial_electric_hourly_loads,
                        initial_hw_hourly_loads,
                        simulation.end_year - simulation.start_year,
                        simulation_output_directory,
                        output,
                        simulation_number,
                        system_performance_outputs,
                        total_loads,
                        total_solar_data[solar.SolarDataType.ELECTRICITY.value]
                        * minigrid.pv_panel.pv_unit,
                    )

                # Carry out an appraisal of the system.
                if electric_yearly_load_statistics is None:
                    raise InternalError(
                        "No electric yearly load statistics were computed for the "
                        "system despite these being needed to appraise the system."
                    )
                system_appraisal: Optional[SystemAppraisal] = appraise_system(
                    electric_yearly_load_statistics,
                    simulation.end_year,
                    finance_inputs,
                    ghg_inputs,
                    location,
                    logger,
                    None,
                    scenario,
                    system_performance_outputs,
                    simulation.start_year,
                    system_details,
                )
            else:
                system_appraisal = None
                logger.info("No analysis to be carried out.")

            # Save the simulation output.
            save_simulation(
                disable_tqdm,
                key_results,
                logger,
                output,
                simulation_output_directory,
                system_performance_outputs,
                simulation_number,
                system_appraisal,
                system_details,
            )

        print(f"Beginning CLOVER simulation runs {'.' * 30}    {DONE}")

        print(
            f"Time taken for simulations: {', '.join(simulation_times)}",
            end="\n",
        )

    if operating_mode == OperatingMode.OPTIMISATION:
        print(f"Beginning CLOVER optimisation runs {'.' * 28}    ", end="\n")
        optimisation_times: List[str] = []

        # Enforce that the optimisation inputs are set correctly before attempting an
        # optimisation.

        if optimisation_inputs is None:
            raise InputFileError(
                "optimisation inputs",
                "Optimisation inputs were not specified despite an optimisation being "
                "called.",
            )
        if electric_yearly_load_statistics is None:
            raise InternalError(
                "Electric yearly load statistics were not correctly computed despite "
                "being needed for an optimisation."
            )

        if (
            optimisation_inputs.number_of_iterations
            * optimisation_inputs.iteration_length
            > location.max_years
        ):
            raise InputFileError(
                "optimisation inputs",
                "An optimisation was requested that runs over the maximum lifetime of "
                "the system.",
            )

        for optimisation_number, optimisation in enumerate(
            tqdm(
                optimisations,
                desc="optimisations",
                disable=disable_tqdm,
                unit="optimisation",
            ),
            1,
        ):
            # Determine the scenario to use for the simulation.
            logger.info("Checking scenario parameters.")
            check_scenario(
                logger, minigrid, operating_mode, parsed_args, optimisation.scenario
            )
            logger.info("Scenario parameters valid.")

            logger.info("Loading grid profile.")
            grid_profile = grid.load_grid_profile(
                auto_generated_files_directory, logger, optimisation.scenario
            )
            logger.info(
                "Grid '%s' profile successfully loaded.",
                optimisation.scenario.grid_type,
            )

            optimisation_string = generate_optimisation_string(
                minigrid, optimisation_inputs, optimisation.scenario
            )
            print(f"Running an optimisation with:\n{optimisation_string}")

            try:
                time_delta, optimisation_results = multiple_optimisation_step(
                    conventional_cw_source_profiles,
                    converters,
                    disable_tqdm,
                    finance_inputs,
                    ghg_inputs,
                    grid_profile,
                    total_solar_data[solar.SolarDataType.TOTAL_IRRADIANCE.value],
                    kerosene_usage,
                    location,
                    logger,
                    minigrid,
                    optimisation,
                    optimisation_inputs,
                    total_solar_data[solar.SolarDataType.TEMPERATURE.value],
                    total_loads,
                    total_solar_data[solar.SolarDataType.ELECTRICITY.value]
                    * minigrid.pv_panel.pv_unit,
                    total_wind_data[wind.WindDataType.WIND_SPEED.value]
                    if total_wind_data is not None
                    else None,
                    electric_yearly_load_statistics,
                )
            except Exception as e:
                print(f"Beginning CLOVER optimisation runs {'.' * 28}    {FAILED}")
                logger.error(
                    "%sAn unexpected error occurred running a CLOVER optimisation. See "
                    "%s for "
                    "details: %s%s",
                    BColours.fail,
                    f"{os.path.join(LOGGER_DIRECTORY, LOGGER_NAME)}.log",
                    str(e),
                    BColours.endc,
                )
                raise

            # Add the time to the counter.
            optimisation_times.append(
                "{0:.3f} s/year".format(  # pylint: disable=consider-using-f-string
                    (time_delta.seconds + time_delta.microseconds * 0.000001)
                    / (
                        optimisation_results[-1].system_details.end_year
                        - optimisation_results[0].system_details.start_year
                    )
                )
            )

            # Add the input file information to the system details file.
            for appraisal in optimisation_results:
                appraisal.system_details.file_information = input_file_info

            # Save the optimisation output.
            save_optimisation(
                disable_tqdm,
                logger,
                optimisation,
                optimisation_inputs,
                optimisation_number,
                output,
                optimisation_output_directory,
                optimisation.scenario,
                optimisation_results,
            )

        print(f"Beginning CLOVER optimisation runs {'.' * 28}    {DONE}")

        print(
            f"Time taken for optimisations: {', '.join(optimisation_times)}",
            end="\n",
        )

    if operating_mode == OperatingMode.PROFILE_GENERATION:
        print("No simulations or optimisations to be carried out.")

    print(
        "Finished. See "
        + os.path.join(LOCATIONS_FOLDER_NAME, parsed_args.location, "outputs")
        + " for output files."
    )


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/python3
########################################################################################
# diesel.py - Energy-system diesel module                                              #
#                                                                                      #
# Author: Ben Winchester                                                               #
# Copyright: Ben Winchester, 2021                                                      #
# License: Open source                                                                 #
# Most recent update: 12/11/2021                                                       #
#                                                                                      #
# For more information, please email:                                                  #
#     philip.sandwell@gmail.com                                                        #
########################################################################################
"""
diesel.py - The energy-system's diesel module.

This module represents diesel generators within the energy system containing
functionality to model diesel generators.

"""

import dataclasses
import logging

from typing import Any, Dict, Optional, Tuple

import numpy as np  # pylint: disable=import-error
import pandas as pd

from ..__utils__ import (
    BColours,
    ELECTRIC_POWER,
    DieselMode,
    InputFileError,
    NAME,
    ProgrammerJudgementFault,
    ResourceType,
)
from ..conversion.conversion import MAXIMUM_OUTPUT, Converter


__all__ = (
    "DIESEL_CONSUMPTION",
    "DieselGenerator",
    "DieselWaterHeater",
    "get_diesel_energy_and_times",
    "get_diesel_fuel_usage",
)


# Diesel consumption:
#   Used to parse diesel fuel consumption information.
DIESEL_CONSUMPTION: str = "diesel_consumption"

# Minimum load:
#   The minimum load that needs to be placed on a diesel-consuming device.
MINIMUM_LOAD: str = "minimum_load"


@dataclasses.dataclass
class DieselGenerator:
    """
    Represents a diesel backup generator.

    .. attribute:: diesel_consumption
        The diesel consumption of the generator, measured in litres per kW produced.

    .. attribute:: minimum_load
        The minimum capacity of the generator, defined between 0 (able to operate with
        any load) and 1 (only able to operate at maximum load).

    .. attribute:: name
        The name of the generator.

    """

    diesel_consumption: float
    minimum_load: float
    name: str


@dataclasses.dataclass
class DieselWaterHeater(Converter):
    """
    Represents a diesel water heater.

    .. attribute:: diesel_consumption
        The diesel consumption of the heater, measured in litres per kWth produced.

    .. attribute:: minimum_load
        The minimum capacity of the heater, defined between 0 (able to operate with any
        load) and 1 (only able to operate at maximum load).

    """

    def __init__(
        self,
        input_resource_consumption: Dict[ResourceType, float],
        maximum_output_capacity: float,
        minimum_load: float,
        name: str,
        output_resource_type: ResourceType,
    ) -> None:
        """
        Instnatiate a :class:`DieselWaterHeater` instance.

        Inputs:
            - input_resource_types:
                The types of load inputted to the d:class:`DieselWaterHeater`evice.
            - maximum_output_capcity:
                The maximum output capacity of the :class:`DieselWaterHeater`.
            - minimum_load:
                The minimum load that must be placed on the :class:`DieselWaterHeater`.
            - name:
                The name of the :class:`DieselWaterHeater`.
            - output_resource_type:
                The type of output produced by the :class:`DieselWaterHeater`.

        """

        # @BenWinchester - Waste consumption needed in this diesel water heater.
        super().__init__(
            input_resource_consumption,
            maximum_output_capacity,
            name,
            output_resource_type,
        )
        self.minimum_load = minimum_load

    @classmethod
    def from_dict(cls, input_data: Dict[str, Any], logger: logging.Logger) -> Any:
        """
        Instantiates a :class:`DieselWaterHeater` instance based on the input data.

        Inputs:
            - input_data:
                The input information, extracted from the diesel inputs YAML file.
            - logger:
                The :class:`logging.Logger` to use for the run.

        Outputs:
            - An instantiated :class:`DieselWaterHeater` based on the input information.

        """

        try:
            input_resource_consumption: Dict[ResourceType, float] = {
                ResourceType.DIESEL: input_data[DIESEL_CONSUMPTION],
                ResourceType.ELECTRIC: input_data[ELECTRIC_POWER]
                if ELECTRIC_POWER in input_data
                else 0,
            }
        except KeyError as e:
            logger.error(
                "%sMissing or invalid information for diesel water heater concerning "
                "its fuel and/or electricity consumption: %s%s",
                BColours.fail,
                str(e),
                BColours.endc,
            )
            raise InputFileError(
                "diesel inputs", "Missing diesel water heater resource inputs."
            ) from None

        return cls(
            input_resource_consumption,
            input_data[MAXIMUM_OUTPUT],
            input_data[MINIMUM_LOAD],
            input_data[NAME],
            ResourceType.HEAT,
        )

    @property
    def diesel_consumption(self) -> float:
        """
        Returns the diesel-fuel consumption of the :class:`DieselWaterHeater`.

        Outputs:
            - The diesel fuel consumption of the :class:`DieselWaterHeater`.

        """

        return self.input_resource_consumption[ResourceType.DIESEL]


def _find_deficit_threshold_blackout(
    unmet_energy: pd.DataFrame, blackouts: pd.DataFrame, backup_threshold: float
) -> Optional[float]:
    """
    Identifies the threshold energy level at which the diesel backup generator turns on
    when the threshold criterion is blackouts.

    Inputs:
        - backup_threshold:
            Desired level of reliability after diesel backup
        - blackouts:
            Current blackout profile before diesel backup
        - unmet_energy:
            Load profile of currently unment energy

    Outputs:
        - energy_threshold:
            The energy threshold (kWh) at which the diesel backup switches on.

    """

    # Find the blackout percentage
    blackout_percentage = float(blackouts.mean(axis=0))  # type: ignore

    # Find the difference in reliability
    reliability_difference = blackout_percentage - backup_threshold
    percentile_threshold = 100.0 * (1.0 - reliability_difference)

    if reliability_difference <= 0.0:
        return None

    return float(np.percentile(unmet_energy, percentile_threshold))


def _find_deficit_threshold_unmet(
    unmet_energy: pd.DataFrame, backup_threshold: float, total_electric_load: float
) -> Optional[float]:
    """
    Identifies the threshold energy level at which the diesel backup generator turns on
    when the threshold criterion is unmet energy.

    Inputs:
        - backup_threshold:
            Desired level of reliability after diesel backup
        - unmet_energy:
            Load profile of currently unment energy
        - total_electric_load:
            The total electric load placed on the system (kWh).

    Outputs:
        - energy_threshold:
            The energy threshold (kWh) at which the diesel backup switches on.

    """

    # Find the blackout percentage
    unmet_energy_percentage = float(np.sum(unmet_energy) / total_electric_load)  # type: ignore

    # Find the difference in reliability
    reliability_difference = unmet_energy_percentage - backup_threshold

    if reliability_difference <= 0:
        return None

    # Sort unmet energy by smallest first
    sorted_unmet_energy = sorted(unmet_energy.values)

    # Loop through hours attributing unmet energy to diesel generator (largest first)
    attributed_unmet_energy: float = 0
    energy_threshold: float

    while attributed_unmet_energy < total_electric_load * reliability_difference:
        energy_threshold = sorted_unmet_energy.pop()
        attributed_unmet_energy += energy_threshold

    return energy_threshold


def _find_deficit_threshold(
    unmet_energy: pd.DataFrame,
    blackouts: pd.DataFrame,
    backup_threshold: float,
    total_electric_load: float,
    diesel_mode: DieselMode,
) -> Optional[float]:
    """
    Identifies the threshold energy level at which the diesel backup generator turns on.

    Inputs:
        - backup_threshold:
            Desired level of reliability after diesel backup
        - blackouts:
            Current blackout profile before diesel backup
        - diesel_mode:
            The diesel mode used in the scenario inputs file
        - unmet_energy:
            Load profile of currently unment energy
        - total_electric_load:
            The total electric load placed on the system (kWh).

    Outputs:
        - energy_threshold:
            The energy threshold (kWh) at which the diesel backup switches on, `None` if
            the diesel generator should not turn on.

    """

    # Find the blackout percentage is mode using blackout criterion
    if diesel_mode == DieselMode.BACKUP:
        return _find_deficit_threshold_blackout(
            unmet_energy, blackouts, backup_threshold
        )

    # Find the blackout percentage is mode using unmet energy criterion
    if diesel_mode == DieselMode.BACKUP_UNMET:
        return _find_deficit_threshold_unmet(
            unmet_energy, backup_threshold, total_electric_load
        )

    raise ProgrammerJudgementFault(
        "src.clover.simulation.diesel::_find_deficit_threshold",
        "Cannot find deficit threshold for non-backup scenarios.",
    )


def get_diesel_energy_and_times(
    unmet_energy: pd.DataFrame,
    blackouts: pd.DataFrame,
    backup_threshold: float,
    total_electric_load: float,
    diesel_mode: DieselMode,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Finds times when the load is greater than the energy threshold.

    Calculates the times at which the diesel backup generator is used, and the energy
    output during those times

    Inputs:
        - backup_threshold:
            Desired level of reliability after diesel backup
        - blackouts:
            Current blackout profile before diesel backup
        - diesel_mode:
            The diesel mode used in the scenario inputs file
        - unmet_energy:
            Load profile of currently unment energy
        - total_electric_load:
            The total electric load placed on the system (kWh).

    Outputs:
        - diesel_energy:
            Profile of energy supplued by diesel backup
        - diesel_times:
            Profile of times when generator is on (1) or off (0)

    """

    energy_threshold = _find_deficit_threshold(
        unmet_energy, blackouts, backup_threshold, total_electric_load, diesel_mode
    )
    if energy_threshold is None:
        return pd.DataFrame([0] * len(unmet_energy)), pd.DataFrame(
            [0] * len(unmet_energy)
        )

    diesel_energy = pd.DataFrame(
        unmet_energy.values * (unmet_energy >= energy_threshold).values
    )
    diesel_times = (unmet_energy >= energy_threshold) * 1
    diesel_times = diesel_times.astype(float)

    return diesel_energy, diesel_times


def get_diesel_fuel_usage(
    capacity: int,
    diesel_generator: DieselGenerator,
    diesel_energy: pd.DataFrame,
    diesel_times: pd.DataFrame,
) -> pd.DataFrame:
    """
    Find diesel fuel consumption.

    Calculates the fuel usage of the diesel backup generator

    Inputs:
        - capacity:
            Capacity (kW) of the diesel generator.
        - diesel_generator:
            The diesel backup generator being modelled.
        - diesel_energy:
            Profile of energy supplued by diesel backup.
        - diesel_times:
            Profile of times when generator is on (1) or off (0).

    Outputs:
        - fuel_usage:
            Hourly profile of diesel fuel usage (litres).

    """

    load_factor: pd.DataFrame = diesel_energy.divide(capacity)  # type: ignore
    above_minimum = pd.DataFrame(
        load_factor.values * (load_factor > diesel_generator.minimum_load).values
    )
    below_minimum = (
        load_factor <= diesel_generator.minimum_load
    ) * diesel_generator.minimum_load
    # @@@ Investigate variable reassignment here.
    load_factor = pd.DataFrame(
        diesel_times.values * (above_minimum.values + below_minimum.values)
    )

    fuel_usage: pd.DataFrame = (
        load_factor * capacity * diesel_generator.diesel_consumption
    )
    fuel_usage = fuel_usage.astype(float)

    return fuel_usage



#!/usr/bin/python3
########################################################################################
# load.py - Load-profile generation module.                                            #
#                                                                                      #
# Authors: Phil Sandwell, Ben Winchester                                               #
# Copyright: Phil Sandwell, 2018                                                       #
# License: Open source                                                                 #
# Most recent update: 14/07/2021                                                       #
#                                                                                      #
# For more information, please email:                                                  #
#     philip.sandwell@gmail.com                                                        #
########################################################################################
"""
load.py - The load-profile generation module of CLOVER.

This module generates load profiles for CLOVER based on the device information passed
in.

"""

import dataclasses
import math
import os

from logging import Logger
from typing import Any, Dict, Optional, Set, Tuple

import numpy as np  # pylint: disable=import-error
import pandas as pd  # pylint: disable=import-error

from tqdm import tqdm  # pylint: disable=import-error

from ..__utils__ import (
    BColours,
    CUT_OFF_TIME,
    ColumnHeader,
    DemandType,
    ELECTRIC_POWER,
    InputFileError,
    InternalError,
    KEROSENE_DEVICE_NAME,
    ResourceType,
    Location,
    Scenario,
    monthly_times_to_daily_times,
)

__all__ = (
    "compute_total_hourly_load",
    "compute_processed_load_profile",
    "DEFAULT_KEROSENE_DEVICE",
    "Device",
    "LOAD_LOGGER_NAME",
    "ResourceType",
    "population_hourly",
    "process_device_hourly_power",
    "process_device_hourly_usage",
    "process_device_ownership",
    "process_device_utilisation",
    "process_load_profiles",
)


# Available:
#   Keyword used for parsing whether a device is available.
AVAILABLE: str = "available"

# Clean-water usage:
#   Keyword used for parsing the clean-water usage of a device.
CLEAN_WATER_USAGE: str = "clean_water_usage"

# Device:
#   Keyword used for parsing the name of a device.
DEVICE: str = "device"

# Final ownership:
#   Keyword used for parsing the final ownership of a device.
FINAL_OWNERSHIP: str = "final_ownership"

# Hot-water usage:
#   Keyword used for parsing the hot-water usage of a device.
HOT_WATER_USAGE: str = "hot_water_usage"

# Initial ownership:
#   Keyword used for parsing the initial ownership of a device.
INITIAL_OWNERSHIP: str = "initial_ownership"

# Innovation:
#   Keyword used for parsing the innovation of a device.
INNOVATION: str = "innovation"

# Imitation:
#   Keyword used for parsing the imitation of a device.
IMITATION: str = "imitation"

# Load logger name:
#   The name to use for the load module logger.
LOAD_LOGGER_NAME: str = "load"

# Mean column name:
#   The name to use for the "mean" column in the yearly-load statistics.
MEAN: str = "Mean"

# Median column name:
#   The name to use for the "median" column in the yearly-load statistics.
MEDIAN: str = "Median"


@dataclasses.dataclass
class Device:
    """
    Represents a device being modelled.

    .. attribute:: available
        Whether the device is available.

    .. attribute:: demand_type
        The type of demand with which the device is associated.

    .. attribute:: electric_power
        The electric power consumption of the device, measured in Watts.

    .. attribute:: final_ownership
        The average ownership of the device per household at the end of the time period
        being modelled.

    .. attribute:: initial_ownership
        The initial average ownership of the device per household.

    .. attribute:: innovation
        The rate of innovation for the device: the rate at which new households acquire
        the device.

    .. attribute:: imitation
        The rate of imitation for the device: the rate at which households copy others
        and acquire the device.

    .. attribute:: name
        The name of the device.

    .. attribute:: clean_water_usage
        The clean-water usage of the device, measured in litres per hour.

    .. attribute:: hot_water_usage
        The hot-water usage of the device, measured in litres per hour.

    """

    available: bool
    demand_type: DemandType
    electric_power: Optional[float]
    final_ownership: float
    initial_ownership: float
    innovation: float
    imitation: float
    name: str
    clean_water_usage: Optional[float]
    hot_water_usage: Optional[float]

    def __hash__(self) -> int:
        """
        Return a unique identifier for the device.

        Outputs:
            A unique identifier for the device.

        """

        return hash(self.__str__())

    def __str__(self) -> str:
        """
        Return a nice-looking output for the device.

        Outputs:
            - A nice-looking string representation for the device.

        """

        representation_string = (
            "Device("
            + f"name={self.name}, "
            + ("available, " if self.available else "unavailable, ")
            + f"electric_power={self.electric_power} W, "
            + f"final_ownership={self.final_ownership}, "
            + f"initial_ownership={self.initial_ownership}, "
            + f"innovation={self.innovation}, "
            + f"imitation={self.imitation}, "
            + f"clean_water_usage={self.clean_water_usage} litres/hour, "
            + f"hot_water_usage={self.hot_water_usage} litres/hour"
            + ")"
        )

        return representation_string

    @classmethod
    def from_dict(cls, device_input: Dict[str, Any]) -> Any:
        """
        Processes input data to generate a :class:`Device` instance.

        Inputs:
            - device_input:
                The device input data extracted from the devices input file.

        Outputs:
            - The :class:`Device` instancce based on the input data.

        """

        demand_type = DemandType(device_input["type"])

        if ELECTRIC_POWER not in device_input:
            raise InputFileError(
                "device inputs",
                "All devices must specify an electric power value explicitly, even if "
                "this is set to zero. "
                f"The device, {device_input['device']}, does not specify such a value.",
            )

        return cls(
            device_input[AVAILABLE],
            demand_type,
            device_input[ELECTRIC_POWER],
            device_input[FINAL_OWNERSHIP],
            device_input[INITIAL_OWNERSHIP],
            device_input[INNOVATION],
            device_input[IMITATION],
            device_input[DEVICE],
            device_input[CLEAN_WATER_USAGE]
            if CLEAN_WATER_USAGE in device_input
            else None,
            device_input[HOT_WATER_USAGE] if HOT_WATER_USAGE in device_input else None,
        )


# Default kerosene device:
#   The default kerosene device to use in the event that no kerosene information is
#   provided.
DEFAULT_KEROSENE_DEVICE = Device(
    False, DemandType.DOMESTIC, 1, 0, 0, 0, 0, KEROSENE_DEVICE_NAME, 0, 0
)


def _cumulative_sales_daily(
    current_market_proportion: float,
    maximum_market_proportion: float,
    innovation: float,
    imitation: float,
    years: int,
) -> pd.DataFrame:
    """
    Return the ownership (sales) of devices in the community.

    Computes and returns the sales (ownership) of devices in the community over the
    lifetime of the simulation.

    Inputs:
        - current_market_proportion:
            The current proportion of potential owners who own the device.
        - maximum_market_proportion:
            The maximum saturation of ownership within the market.
        - innovation:
            The rate of innovation, i.e., new customers spontaneously purchasing the
            device.
        - imitation:
            The rate of customers copying those around them and purchasing the device.
        - years:
            The number of years for which the simulation is being run.

    Outputs:
        - A dataframe of the ownership of that device type for each day.

    Notes:
        Uses the Bass diffusion model

    """

    maximisation_ratio = maximum_market_proportion - current_market_proportion
    daily_innovation = innovation / 365
    daily_imitation = imitation / 365
    cum_sales = {}
    for day in range(0, 365 * years):
        num = 1 - math.exp(-1 * (daily_innovation + daily_imitation) * day)
        den = 1 + (daily_imitation / daily_innovation) * math.exp(
            -1 * (daily_innovation + daily_imitation) * day
        )
        cum_sales[day] = maximisation_ratio * num / den + current_market_proportion
    return pd.DataFrame(list(cum_sales.values()))


def _population_growth_daily(
    community_growth_rate: float, community_size: int, num_years: int
) -> pd.DataFrame:
    """
    Calculates the growth in the number of households in the community

    Inputs:
        Takes inputs from "Location inputs.csv" in the "Location data" folder

    Outputs:
        Gives a DataFrame of the number of households in the community for each day

    Notes:
        Simple compound interest-style growth rate

    """

    population = []
    growth_rate_daily = (1 + community_growth_rate) ** (1 / 365.0) - 1
    for day in range(0, 365 * num_years):
        population.append(math.floor(community_size * (1 + growth_rate_daily) ** day))
    return pd.DataFrame(population)


def _yearly_load_statistics(total_load: pd.DataFrame, years: int) -> pd.DataFrame:
    """
    Calculates the load statistics for each year on an hourly basis.

    Inputs:
        - total_load:
            Hourly total load of the system.
        - years:
            The number of years for which the simulation is being run.

    Outputs:
        - A dataframe containing the maximum, mean and median hourly loads.

    """

    total_load_yearly = pd.DataFrame(
        np.reshape(
            pd.DataFrame(total_load.sum(axis=1)).values,
            (years, 365 * 24),
        )
    )

    yearly_maximum: pd.DataFrame = pd.DataFrame(
        total_load_yearly.max(axis=1)  # type: ignore
    )
    yearly_maximum.columns = pd.Index([ColumnHeader.MAXIMUM.value])
    yearly_mean: pd.DataFrame = pd.DataFrame(
        total_load_yearly.mean(axis=1).round(0)  # type: ignore
    )
    yearly_mean.columns = pd.Index([MEAN])
    yearly_median: pd.DataFrame = pd.DataFrame(
        np.percentile(total_load_yearly, 50, axis=1)  # type: ignore
    )
    yearly_median.columns = pd.Index([MEDIAN])
    yearly_load_statistics = pd.concat(
        [yearly_maximum, yearly_mean, yearly_median], axis=1
    )

    return yearly_load_statistics


def _number_of_devices_daily(
    device: Device,
    location: Location,
    logger: Logger,
) -> pd.DataFrame:
    """
    Calculates the number of devices owned by the community on each day

    Inputs:
        - device:
            The device currently being considered.
        - location:
            The location currently being considered.
        - logger:
            The logger to use for the run.
    Outputs:
        - daily_ownership:
            Returns the number of devives that are owned by the community on a given
            day. Devices which are not permitted by "Devices.csv" should return a list
            composed entirely of zeroes.

    """

    if device.available:
        logger.info(
            "Calculating ownership for device %s.",
            device.name,
        )
        population_growth_rate = _population_growth_daily(
            location.community_growth_rate,
            location.community_size,
            location.max_years,
        )
        if device.final_ownership != device.initial_ownership:
            logger.info(
                "%s ownership changes over time, calculating.",
                device.name,
            )
            cum_sales = _cumulative_sales_daily(
                device.initial_ownership,
                device.final_ownership,
                device.innovation,
                device.imitation,
                location.max_years,
            )
            daily_ownership = pd.DataFrame(
                np.floor(cum_sales.values * population_growth_rate.values)
            )

        else:
            logger.info(
                "%s ownership remains constant.",
                device.name,
            )
            daily_ownership = pd.DataFrame(
                np.floor(population_growth_rate * device.initial_ownership)  # type: ignore
            )
        logger.info(
            "Ownership for device %s calculated.",
            device.name,
        )

    else:
        logger.info(
            "Device %s was marked as unavailable, setting ownership to zero.",
            device.name,
        )
        daily_ownership = pd.DataFrame(np.zeros((location.max_years * 365, 1)))

    return daily_ownership


def compute_total_hourly_load(  # pylint: disable=too-many-locals
    *,
    device_hourly_loads: Dict[str, pd.DataFrame],
    devices: Set[Device],
    disable_tqdm: bool,
    generated_device_load_filepath: str,
    logger: Logger,
    total_load_profile: Optional[pd.DataFrame],
    years: int,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Calculates the aggregated load of all devices.

    Inputs:
        - device_hourly_loads:
            A mapping between device name and the hourly load profile of the device.
        - devices:
            The set of devices included in the system.
        - disable_tqdm:
            Whether to disable the tqdm progress bars (True) or display them (False).
        - generated_device_load_filepath:
            The directory in which to store the generated hourly load profiles for the
            device.
        - logger:
            The logger to use for the run.
        - regenerate:
            Whether to force regenerate the profiles.
        - total_load_profile:
            The total load profile to use in lieu of profile generation if specified.
        - years:
            The nbumber of years for which the simulation is being run.

    Outputs:
        - The total load of all devices;
        - The yearly-load statistics.

    """

    # If the files already exist, simply read the data and quit.
    total_load_filepath = os.path.join(generated_device_load_filepath, "total_load.csv")
    yearly_load_statistics_filepath = os.path.join(
        generated_device_load_filepath, "yearly_load_statistics.csv"
    )

    if total_load_profile is None:
        logger.info(
            "Total load data must be recomputed each run unless an override profile is "
            "specified. Calculating total load data."
        )

        # Instantiate empty dataframes.
        domestic_load = pd.DataFrame(np.zeros((years * 365 * 24, 1)))
        commercial_load = pd.DataFrame(np.zeros((years * 365 * 24, 1)))
        public_load = pd.DataFrame(np.zeros((years * 365 * 24, 1)))

        # Sum over the device loads.
        for device in tqdm(
            devices,
            desc="total load profile",
            disable=disable_tqdm,
            leave=True,
            unit="device",
        ):
            # Skip the device if it is not available in the community.
            if not device.available:
                continue

            if device.demand_type == DemandType.DOMESTIC:
                domestic_load = pd.DataFrame(
                    domestic_load.values + device_hourly_loads[device.name].values
                )
            elif device.demand_type == DemandType.COMMERCIAL:
                commercial_load = pd.DataFrame(
                    commercial_load.values + device_hourly_loads[device.name].values
                )
            elif device.demand_type == DemandType.PUBLIC:
                public_load = pd.DataFrame(
                    public_load.values + device_hourly_loads[device.name].values
                )
            else:
                logger.error(
                    "Demand type of device %s is unknown. Type: %s.",
                    device.name,
                    device.demand_type,
                )

        logger.info("Total load for all devices successfully computed.")
        total_load = pd.concat([domestic_load, commercial_load, public_load], axis=1)
        total_load.columns = pd.Index(
            [
                DemandType.DOMESTIC.value,
                DemandType.COMMERCIAL.value,
                DemandType.PUBLIC.value,
            ]
        )

        logger.info("Saving total load.")
        with open(total_load_filepath, "w") as f:
            total_load.to_csv(f, line_terminator="")  # type: ignore
        logger.info("Total device load successfully saved to %s.", total_load_filepath)

    else:
        total_load = total_load_profile
        if not all(
            total_load.columns
            == pd.Index(
                [
                    DemandType.DOMESTIC.value,
                    DemandType.COMMERCIAL.value,
                    DemandType.PUBLIC.value,
                ]
            )
        ):
            logger.error(
                "%sThe total load profile specified is not of the right format. See "
                "logs for details.%s",
                BColours.fail,
                BColours.endc,
            )
            logger.info(
                "The total load file given must have columns which match %s.",
                ", ".join(f"'{e.value}'" for e in DemandType),
            )
            raise InputFileError(
                "total-load file",
                "The total load profile is not of the correct format.",
            )

    # Attempt to read the yearly load statistics from a file and compute if it doesn't
    # exist.
    logger.info(
        "Yearly load statistics must be recomputed, calculating yearly load statistics."
    )
    yearly_load_statistics = _yearly_load_statistics(total_load, years)

    logger.info("Saving yearly load statistics.")
    with open(yearly_load_statistics_filepath, "w") as f:
        yearly_load_statistics.to_csv(
            f, index=False, line_terminator=""  # type: ignore
        )
    logger.info("Yearly load statistics successfully saved.")

    return total_load, yearly_load_statistics


def population_hourly(location: Location) -> pd.DataFrame:
    """
    Calculates the growth in the number of households in the community for each hour

    Inputs:
        - location:
            The location being considered.

    Outputs:
        - A DataFrame of the number of households in the community for each hour.

    Notes:
        Simple compound interest-style growth rate.

    """

    growth_rate_hourly = (1 + location.community_growth_rate) ** (
        1 / (24.0 * 365.0)
    ) - 1
    population = [
        math.floor(location.community_size * (1 + growth_rate_hourly) ** hour)
        for hour in range(0, 365 * 24 * location.max_years)
    ]

    return pd.DataFrame(population)


def process_device_hourly_power(
    device: Device,
    *,
    generated_device_load_filepath: str,
    hourly_device_usage: pd.DataFrame,
    resource_type: ResourceType,
    logger: Logger,
    regenerate: bool,
) -> pd.DataFrame:
    """
    Calculate the hourly power consumption of the device.

    Inputs:
        - device:
            The device to be processed.
        - generated_device_load_filepath:
            The directory in which to store the generated hourly load profiles for the
            device.
        - hourly_device_usage:
            The hourly usage profile for the device.
        - logger:
            The logger to use for the run.
        - power_type:
            The type of power being investigated, e.g., "electric_power".
        - regenerate:
            Whether to force-regenerate the various profiles.

    Outputs:
        - The hourly load of the device as a :class:`pandas.DataFrame`.

    """

    filename = f"{device.name}_load.csv"
    hourly_usage_filepath = os.path.join(generated_device_load_filepath, filename)

    # If the hourly power usage file already exists, load the data in.
    logger.info("Processing hourly power profile for %s.", device.name)
    if os.path.isfile(hourly_usage_filepath) and not regenerate:
        with open(hourly_usage_filepath, "r") as f:
            device_load: pd.DataFrame = pd.read_csv(f, header=None)
        logger.info(
            "Hourly power profile for %s successfully read from file %s.",
            device.name,
            hourly_usage_filepath,
        )
    else:
        # Compute the hourly load profile.
        logger.info("Computing hourly power usage for %s.", device.name)
        if resource_type == ResourceType.ELECTRIC:
            if device.electric_power is None:
                raise Exception(
                    f"{BColours.fail}Internal error processing device "
                    + f"'{device.name}', electric power unexpectedly `None`.{BColours.endc}",
                )
            device_load = hourly_device_usage * device.electric_power
            logger.info(
                "Electric hourly power usage for %s successfully computed.", device.name
            )
        elif resource_type == ResourceType.CLEAN_WATER:
            if device.clean_water_usage is None:
                raise InternalError(
                    f"{BColours.fail}Internal error processing device "
                    + f"'{device.name}', clean-water usage unexpectedly `None`."
                    + f"{BColours.endc}",
                )

            device_load = hourly_device_usage * device.clean_water_usage
            logger.info("Water usage for %s successfully computed.", device.name)
        elif resource_type == ResourceType.HOT_CLEAN_WATER:
            if device.hot_water_usage is None:
                raise InternalError(
                    f"{BColours.fail}Internal error processing device "
                    + f"'{device.name}', hot-water usage unexpectedly `None`."
                    + f"{BColours.endc}",
                )

            device_load = hourly_device_usage * device.hot_water_usage
        else:
            logger.error(
                "%sUnsuported load type used: %s%s",
                BColours.fail,
                resource_type.value,
                BColours.endc,
            )

        # Reset the index on the device load.
        try:
            device_load = device_load.reset_index(drop=True)
        except UnboundLocalError:
            logger.error(
                "%sHandling of resource load failed due to unknown load type: %s%s",
                BColours.fail,
                resource_type.value,
                BColours.endc,
            )
            raise

        # Save the hourly power profile.
        logger.info("Saving hourly power usage for %s.", device.name)

        with open(
            hourly_usage_filepath,
            "w",
        ) as f:
            device_load.to_csv(
                f, header=None, index=False, line_terminator=""  # type: ignore
            )

        logger.info(
            "Hourly power proifle for %s successfully saved to %s.",
            device.name,
            hourly_usage_filepath,
        )

    return device_load


def process_device_hourly_usage(
    device: Device,
    *,
    daily_device_ownership: pd.DataFrame,
    daily_device_utilisation: pd.DataFrame,
    generated_device_usage_filepath: str,
    logger: Logger,
    regenerate: bool,
    years: int,
) -> pd.DataFrame:
    """
    Calculate the number of devices in use by the community.

    Inputs:
        - device:
            The data for the device to be processed.
        - daily_device_ownership:
            The ownership data for the device.
        - daily_device_utilisation:
            The utilisation data for the device.
        - generated_device_usage_filepath:
            The directory in which to store the generated hourly device-usage profiles.
        - logger:
            The logger to use for the run.
        - regenerate:
            Whether to force-regenerate the profiles.
        - years:
            The number of years for which the simulation is being run.

    Outputs:
        - The hourly usage of the device specified as a :class:`pandas.DataFrame`.

    """

    filename = f"{device.name}_in_use.csv"
    filepath = os.path.join(generated_device_usage_filepath, filename)

    # If the device hourly usage already exists, then read the data in from the file.
    if os.path.isfile(filepath) and not regenerate:
        with open(filepath, "r") as f:
            hourly_device_usage = pd.read_csv(f, header=None)
        logger.info(
            "Hourly device usage for %s successfully read from file: %s",
            device.name,
            filepath,
        )

    else:
        daily_device_utilisation = daily_device_utilisation.reset_index(drop=True)
        daily_device_ownership = daily_device_ownership.reset_index(drop=True)

        hourly_device_usage = pd.DataFrame()

        # Calculate the hourly-usage profile.
        # This processes a random distribution for usage based on the device ownership and
        # utilisation on any given day for all days within the simulation range.
        #

        logger.info("Calculating number of %ss in use", device.name)
        try:
            hourly_device_usage = pd.concat(  # type: ignore
                [
                    pd.DataFrame(  # type: ignore
                        np.random.binomial(  # type: ignore
                            float(daily_device_ownership.iloc[day, 0]),
                            daily_device_utilisation.iloc[day, :],
                        )
                    )
                    for day in range(0, 365 * years)
                ]
            )
        except ValueError as e:
            logger.error(
                "%sError computing device hourly usage profile for %s: type error in "
                "variables: %s%s",
                BColours.fail,
                device.name,
                str(e),
                BColours.endc,
            )
            raise

        logger.info("Hourly usage profile for %s successfully calculated.", device.name)

        # Save the hourly-usage profile.
        logger.info("Saving hourly usage profile for %s.", device.name)

        with open(
            filepath,
            "w",
        ) as f:
            hourly_device_usage.to_csv(
                f, header=None, index=False, line_terminator=""  # type: ignore
            )

        logger.info(
            "Hourly usage proifle for %s successfully saved to %s.",
            device.name,
            filename,
        )

    return hourly_device_usage


def process_device_ownership(
    device: Device,
    *,
    generated_device_ownership_directory: str,
    location: Location,
    logger: Logger,
    regenerate: bool,
) -> pd.DataFrame:
    """
    Process device-data files.

    Processes the device files, including device ownership and utilisation, for a given
    device.

    Inputs:
        - device:
            The device to be processed.
        - generated_device_ownership_directory:
            The directory in which to store the generated device-ownership profiles.
        - location:
            The location currently being considered.
        - logger:
            The logger to use for the run.
        - regenerate:
            Whether to force-regenerate the profiles.

    Outputs:
        - The daily ownership of the device.

    """

    logger.info("Load ownership process instantiated for device %s.", device.name)

    daily_ownership_filename = f"{device.name}_daily_ownership.csv"
    daily_ownership_filepath = os.path.join(
        generated_device_ownership_directory,
        daily_ownership_filename,
    )

    # If the daily ownership file already exists, then read the data from the file.
    if os.path.isfile(daily_ownership_filepath) and not regenerate:
        with open(daily_ownership_filepath, "r") as f:
            daily_ownership = pd.read_csv(f, header=None)
        logger.info(
            "Monthly device-ownership profile for %s successfully read from %s.",
            device.name,
            daily_ownership_filepath,
        )

    else:
        # Compute the device-ownership profiles.
        logger.info("Computing device ownership for %s.", device.name)

        # Compute the daily device usage.
        daily_ownership = _number_of_devices_daily(
            device,
            location,
            logger,
        )
        logger.info(
            "Monthly device ownership profile for %s successfully computed.",
            device.name,
        )

        # Save the usage to the output file.
        with open(
            daily_ownership_filepath,
            "w",
        ) as f:
            daily_ownership.to_csv(
                f, header=None, index=False, line_terminator=""  # type: ignore
            )
        logger.info(
            "Monthly deivice-ownership profile for %s successfully saved to %s.",
            device.name,
            daily_ownership_filename,
        )

    return daily_ownership


def process_device_utilisation(
    device: Device,
    *,
    device_utilisations: Dict[Device, pd.DataFrame],
    generated_device_utilisation_directory: str,
    location: Location,
    logger: Logger,
    regenerate: bool,
) -> pd.DataFrame:
    """
    Process device-data files.

    Processes the device files, including device ownership and utilisation, for a given
    device.

    Inputs:
        - device:
            The data for the device to be processed.
        - device_uilisations:
            The set of device utilisations to process.
        - generated_device_utilisation_directory:
            The directory in which to store the generated device-utilisation profiles.
        - location:
            The location currently being considered.
        - logger:
            The logger to use for the run.
        - regenerate:
            Whether to force-regenerate the profiles.

    Outputs:
        - The interpolated daily utilisation profile.

    """

    logger.info("Load utilisation process instantiated for device %s.", device.name)

    daily_times_filename = f"{device.name}_daily_times.csv"
    filepath = os.path.join(
        generated_device_utilisation_directory, daily_times_filename
    )

    # If the file already exists, simply read in the data.
    if os.path.isfile(filepath) and not regenerate:
        with open(filepath, "r") as f:
            interpolated_daily_profile = pd.read_csv(f, header=None)
        logger.info(
            "Daily device-utilisation profile for %s successfully read from file %s.",
            device.name,
            filepath,
        )

    else:
        logger.info("Computing device-utilisation profile for %s.", device.name)
        interpolated_daily_profile = monthly_times_to_daily_times(
            device_utilisations[device],
            location.max_years,
        )
        logger.info(
            "Daily device-utilisation profile for %s successfully computed.",
            device.name,
        )

        # Save this to the output file.
        with open(filepath, "w") as f:
            interpolated_daily_profile.to_csv(
                f, header=None, index=False, line_terminator=""  # type: ignore
            )
        logger.info(
            "Daily deivice-utilisation profile for %s successfully saved to %s.",
            device.name,
            daily_times_filename,
        )

    return interpolated_daily_profile


def process_load_profiles(  # pylint: disable=too-many-locals
    auto_generated_files_directory: str,
    device_utilisations: Dict[Device, pd.DataFrame],
    disable_tqdm: bool,
    location: Location,
    logger: Logger,
    regenerate: bool,
    resource_type: ResourceType,
    total_load_profile: Optional[pd.DataFrame] = None,
) -> Tuple[Dict[str, pd.DataFrame], pd.DataFrame, pd.DataFrame]:
    """
    Process all the load information and profiles to generate the total load.

    This function runs through the processing of all the various load profiles and
    is the default access point for this module.

    Inputs:
        - auto_generated_files_directory:
            The directory in which auto-generated files should be saved.
        - device_utilisations:
            The processed device utilisation information.
        - disable_tqdm:
            Whether to disable the tqdm progress bars (True) or display them (False).
        - resource_type:
            The type of load being investigated.
        - location:
            The location currently being considered.
        - logger:
            The logger to use for the run.
        - regenerate:
            Whether to force-regenerate the various profiles.
        - total_load_profile:
            The total load profile to use in lieu of profile generation if specified.

    Outputs:
        - A mapping between the device and the load demand that it generates for the
          first few cycles;
        - The total load use for all devices;
        - The yearly load statistics.

    """

    device_hourly_loads: Dict[str, pd.DataFrame] = {}
    if resource_type == ResourceType.ELECTRIC:
        resource_name: str = "electric"
        relevant_device_utilisations: Dict[Device, pd.DataFrame] = {
            device: device_utilisation
            for device, device_utilisation in device_utilisations.items()
            if device.electric_power is not None
        }
    elif resource_type == ResourceType.CLEAN_WATER:
        resource_name = "clean_water"
        relevant_device_utilisations = {
            device: device_utilisation
            for device, device_utilisation in device_utilisations.items()
            if device.clean_water_usage is not None
        }
    elif resource_type == ResourceType.HOT_CLEAN_WATER:
        resource_name = "hot_water"
        relevant_device_utilisations = {
            device: device_utilisation
            for device, device_utilisation in device_utilisations.items()
            if device.hot_water_usage is not None
        }

    else:
        logger.error(
            "%sUnknown load type when calling the load module to generate profiles: %s%s",
            BColours.fail,
            resource_type.value,
            BColours.endc,
        )
        raise Exception(
            f"{BColours.fail}Unknown load type: {resource_type.value}{BColours.endc}"
        )

    for device in tqdm(
        relevant_device_utilisations,
        desc=f"{resource_name.replace('_', ' ')} load profiles",
        disable=disable_tqdm,
        leave=True,
        unit="device",
    ):
        # If the device is not available, then skip it.
        if not device.available and not device.name == KEROSENE_DEVICE_NAME:
            continue

        # Compute the device ownership.
        daily_device_ownership = process_device_ownership(
            device,
            generated_device_ownership_directory=os.path.join(
                auto_generated_files_directory,
                "load",
                "device_ownership",
            ),
            location=location,
            logger=logger,
            regenerate=regenerate,
        )
        logger.info(
            "Device ownership information for %s successfully computed.",
            device.name,
        )

        # Compute the device utilisation.
        daily_device_utilisaion = process_device_utilisation(
            device,
            device_utilisations=relevant_device_utilisations,
            generated_device_utilisation_directory=os.path.join(
                auto_generated_files_directory,
                "load",
                "device_utilisation",
            ),
            location=location,
            logger=logger,
            regenerate=regenerate,
        )
        logger.info(
            "Device utilisation information for %s successfully computed.",
            device.name,
        )

        # Compute the device usage.
        hourly_device_usage = process_device_hourly_usage(
            device,
            daily_device_ownership=daily_device_ownership,
            daily_device_utilisation=daily_device_utilisaion,
            generated_device_usage_filepath=os.path.join(
                auto_generated_files_directory, "load", "device_usage"
            ),
            logger=logger,
            years=location.max_years,
            regenerate=regenerate,
        )
        logger.info(
            "Device hourly usage information for %s successfully computed.",
            device.name,
        )

        # Compute the load profile based on this usage.
        device_hourly_loads[device.name] = process_device_hourly_power(
            device,
            generated_device_load_filepath=os.path.join(
                auto_generated_files_directory, "load", resource_name, "device_load"
            ),
            hourly_device_usage=hourly_device_usage,
            resource_type=resource_type,
            logger=logger,
            regenerate=regenerate,
        )
        logger.info(
            "Device hourly load information for %s successfully computed.",
            device.name,
        )

    logger.info("Computing the total device hourly load and yearly load statistics.")
    total_load, yearly_statistics = compute_total_hourly_load(
        device_hourly_loads=device_hourly_loads,
        devices=set(relevant_device_utilisations.keys()),
        disable_tqdm=disable_tqdm,
        generated_device_load_filepath=os.path.join(
            auto_generated_files_directory, "load", resource_name, "device_load"
        ),
        logger=logger,
        total_load_profile=total_load_profile,
        years=location.max_years,
    )
    logger.info("Total load and yearly statistics successfully computed.")

    return (
        {
            device_name: load.iloc[0:CUT_OFF_TIME, :]
            for device_name, load in device_hourly_loads.items()
        },
        total_load,
        yearly_statistics,
    )


def compute_processed_load_profile(
    scenario: Scenario, total_load: pd.DataFrame
) -> pd.DataFrame:
    """
    Gets the total community load over 20 years in kW

    Inputs:
        - scenario:
            Information about the scenario currently being run.
        - total_load:
            The total load as a :class:`pandas.DataFrame`.

    Outputs:
        - A :class:`pandas.DataFrame` with columns for the load of domestic,
            commercial and public devices.

    """

    processed_total_load: Optional[pd.DataFrame] = None

    if scenario.demands.domestic:
        processed_total_load = pd.DataFrame(
            total_load[DemandType.DOMESTIC.value].values
        )

    if scenario.demands.commercial:
        if processed_total_load is not None:
            processed_total_load += pd.DataFrame(  # type: ignore
                total_load[DemandType.COMMERCIAL.value].values
            )
        else:
            processed_total_load = pd.DataFrame(total_load[DemandType.COMMERCIAL.value])

    if scenario.demands.public:
        if processed_total_load is not None:
            processed_total_load += pd.DataFrame(  # type: ignore
                total_load[DemandType.PUBLIC.value].values
            )
        else:
            processed_total_load = pd.DataFrame(total_load[DemandType.PUBLIC.value])

    if processed_total_load is None:
        raise Exception("At least one load type must be specified.")

    return processed_total_load



#!/usr/bin/python3
########################################################################################
# hpc.py - Wrapper script around CLOVER when run on Imperial College's HPC.            #
#                                                                                      #
# Authors: Phil Sandwell, Ben Winchester                                               #
# Copyright: Phil Sandwell, 2022                                                       #
# Date created: 29/03/2022                                                             #
# License: Open source                                                                 #
#                                                                                      #
# For more information, please email:                                                  #
#   benedict.winchester@gmail.com                                                      #
#   philip.sandwell@gmail.com                                                          #
########################################################################################
"""
hpc.py - The wrapper script for running CLOVER on the HPC.

Imperial College London owns a series of high-performance computers. This module
provides a wrapper around the main functionality of CLOVER to enable it to be run on
the HPC.

"""

import os
import sys
from typing import Any, List

from ..__main__ import main as clover_main
from ..__utils__ import BColours, InternalError, get_logger
from .hpc_utils import (
    HpcOptimisation,
    HpcRunType,
    HpcSimulation,
    InvalidRunError,
    parse_args_and_hpc_input_file,
    temporary_optimisations_file,
)


__all__ = ("main",)

# HPC Job Number:
#   Name of the environment variable for the HPC job number.
HPC_JOB_NUMBER: str = "PBS_ARRAY_INDEX"


# Logger name:
#   The name to use for the logger for this script.
LOGGER_NAME: str = "hpc_run_{}"


def main(args: List[Any]) -> None:
    """
    Wrapper around CLOVER when run on the HPC.

    """

    # Determine the run that is to be carried out.
    try:
        hpc_job_number = int(os.getenv(HPC_JOB_NUMBER))  # type: ignore
    except ValueError:
        print(
            f"{BColours.fail}HPC environmental variable {HPC_JOB_NUMBER} was not of "
            + f"type int.{BColours.endc}",
        )
        raise
    logger = get_logger(LOGGER_NAME.format(hpc_job_number), False)
    logger.info("HPC run script executed.")
    logger.info("CLI arguments: %s", ", ".join(args))

    # Call the utility module to parse the HPC run information.
    logger.info("Parsing HPC input file.")
    _, runs, verbose, _ = parse_args_and_hpc_input_file(args, logger)
    logger.info("HPC input file successfully parsed.")

    # Determine the run.
    run_number: int = hpc_job_number - 1
    try:
        hpc_run = runs[run_number]
    except IndexError:
        logger.error(
            "%sRun number %s out of bounds. Only %s runs submitted.%s",
            BColours.fail,
            hpc_job_number,
            len(runs),
            BColours.endc,
        )
        raise
    logger.info("Run successfully determined: %s", str(hpc_run))

    # Setup the arguments to pass to CLOVER.
    clover_arguments = [
        "--location",
        hpc_run.location,
        "--output",
        f"{hpc_run.output}_hpc_run_{run_number}",
    ]

    if hpc_run.total_load:
        if hpc_run.total_load_file is None:
            logger.error(
                "%sRun %s was processed as having a total-load file but an internal "
                "error occurred determining the total-load file name.%s",
                BColours.fail,
                hpc_job_number,
                BColours.endc,
            )
            raise InternalError(
                "Error occurred processing total-load filename for run #"
                f"{hpc_job_number}."
            )
        clover_arguments.extend(["--electric-load-profile", hpc_run.total_load_file])

    if verbose:
        clover_arguments.append("--verbose")

    if hpc_run.type == HpcRunType.OPTIMISATION:
        if not isinstance(hpc_run, HpcOptimisation):
            logger.error(
                "%sRun %s marked as optimisation but was processed as a simulation.%s",
                BColours.fail,
                hpc_job_number,
                BColours.endc,
            )
            raise InternalError(
                f"Failure processing run #{hpc_job_number} as a optimisation."
            )

        logger.info("Run %s is an optimisation.", hpc_job_number)

        # Run the optimisation with the temporary optimisations file.
        with temporary_optimisations_file(
            logger, hpc_run, run_number
        ) as optimisation_inputs_filename:
            clover_arguments.extend(
                [
                    "--optimisation",
                    "--optimisation-inputs-file",
                    optimisation_inputs_filename,
                ]
            )
            logger.info(
                "Temporary optimisation file successfully created: %s",
                optimisation_inputs_filename,
            )

            # Call CLOVER with this information.
            logger.info("Calling CLOVER with arguments: %s", " ".join(clover_arguments))
            clover_main(clover_arguments, True, hpc_job_number)

    elif hpc_run.type == HpcRunType.SIMULATION:
        if not isinstance(hpc_run, HpcSimulation):
            logger.error(
                "%sRun %s marked as simulation but was processed as an optimisation.%s",
                BColours.fail,
                hpc_job_number,
                BColours.endc,
            )
            raise InternalError(
                f"Failure processing run #{hpc_job_number} as a simulation."
            )

        logger.info("Run %s is a simulation.", hpc_job_number)
        clover_arguments.extend(
            [
                "--simulation",
                "--pv-system-size",
                str(hpc_run.pv_system_size),
                "--storage-size",
                str(hpc_run.storage_size),
            ]
        )

        # Call CLOVER with this information.
        logger.info("Calling CLOVER with arguments: %s", " ".join(clover_arguments))
        clover_main(clover_arguments, True, hpc_job_number)

    else:
        logger.error(
            "%sRun %s was not a supported run type. Supported run types are %s.%s",
            BColours.fail,
            hpc_job_number,
            ", ".join(str(e.value) for e in HpcRunType),
            BColours.endc,
        )
        raise InvalidRunError(f"Run {hpc_job_number} was not of a supported run type.")


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/python3
########################################################################################
# __init__.py - Python internals module, used to expose code here.                     #
#                                                                                      #
# Author: Ben Winchester                                                               #
# Copyright: Ben Winchester, 2021                                                      #
# Date created: 05/07/2021                                                             #
# License: Open source                                                                 #
########################################################################################
"""
The internals module"""

from .__main__ import main
from .__utils__ import (
    BColours,
    CleanWaterMode,
    Criterion,
    CUT_OFF_TIME,
    daily_sum_to_monthly_sum,
    DEFAULT_SCENARIO,
    DemandType,
    DesalinationScenario,
    dict_to_dataframe,
    DieselMode,
    DONE,
    ELECTRIC_POWER,
    EXCHANGER,
    FAILED,
    get_logger,
    HEAT_CAPACITY_OF_WATER,
    HotWaterScenario,
    hourly_profile_to_daily_sum,
    HTFMode,
    InputFileError,
    InternalError,
    KEROSENE_DEVICE_NAME,
    KeyResults,
    Location,
    LOCATIONS_FOLDER_NAME,
    LOGGER_DIRECTORY,
    monthly_profile_to_daily_profile,
    monthly_times_to_daily_times,
    NAME,
    open_simulation,
    OperatingMode,
    PACKAGE_NAME,
    ProgrammerJudgementFault,
    RAW_CLOVER_PATH,
    read_yaml,
    RenewableEnergySource,
    ResourceType,
    RenewablesNinjaError,
    save_simulation,
    Scenario,
    Simulation,
    SystemAppraisal,
    SystemDetails,
    ZERO_CELCIUS_OFFSET,
)
from .analysis import *
from .argparser import parse_args, validate_args
from .fileparser import (
    GENERATION_INPUTS_FILE,
    INPUTS_DIRECTORY,
    KEROSENE_TIMES_FILE,
    KEROSENE_USAGE_FILE,
    parse_input_files,
    parse_scenario_inputs,
)
from .printer import generate_optimisation_string, generate_simulation_string



#!/usr/bin/python3
########################################################################################
# argparser.py - Argument-parsing code for CLOVER.                                     #
#                                                                                      #
# Author: Ben Winchester                                                               #
# Copyright: Ben Winchester, 2021                                                      #
# Date created: 13/07/2021                                                             #
# License: Open source                                                                 #
########################################################################################
"""
argparser.py - The argument-parsing module for CLOVER.

"""

import argparse
import logging

from typing import Any, List

from .__utils__ import BColours, DEFAULT_SCENARIO

__all__ = (
    "parse_args",
    "validate_args",
)


class MissingParametersError(Exception):
    """
    Raised when not all parameters have been specified on the command line.

    """

    def __init__(self, missing_parameter: str) -> None:
        """
        Instantiate a missing parameters error.

        Inputs:
            - missing_parameter:
                The parameter which has not been specified.

        """

        super().__init__(
            f"Missing command-line parameters: {missing_parameter}. "
            + "Run `clover --help` for more information."
        )


def parse_args(args: List[Any]) -> argparse.Namespace:
    """
    Parses command-line arguments into a :class:`argparse.NameSpace`.

    """

    parser = argparse.ArgumentParser()

    # Misc. arguments.
    # Fast mode:
    #   Used for debugging purposes to run with fsat models.
    parser.add_argument(
        "--debug", "-d", action="store_true", default=False, help=argparse.SUPPRESS
    )
    # Verbose:
    #   Used for generating verbose logs for debugging.
    parser.add_argument(
        "--verbose", "-v", action="store_true", default=False, help=argparse.SUPPRESS
    )

    # Mandatory arguments regardless of the use case.
    mandatory_arguments = parser.add_argument_group("mandatory arguments")
    mandatory_arguments.add_argument(
        "--location",
        "-l",
        type=str,
        help="The name of the location for which to run CLOVER.",
    )

    # Arguments used only when specifying how profiles should be generated or used.
    profile_arguments = parser.add_argument_group("profile arguments")
    profile_arguments.add_argument(
        "--refetch",
        action="store_true",
        default=False,
        help="If specified, CLOVER will refetch the various profiles used from the "
        "renewables.ninja API. Otherwise, existing profiles will be used if present.",
    )
    profile_arguments.add_argument(
        "--regenerate",
        action="store_true",
        default=False,
        help="If specified, CLOVER will regenerate the various profiles used. "
        "Otherwise, existing profiles will be used if present.",
    )

    # Arguments in common to both simulations and optimisations
    action_arguments = parser.add_argument_group(
        "simulation and optimisation arguments",
    )
    action_arguments.add_argument(
        "--electric-load-profile",
        "-el",
        type=str,
        help="The name of the load profile to use for the run. This overrides CLOVER's "
        "in-built load-profile generation.",
    )
    action_arguments.add_argument(
        "--output",
        "-o",
        type=str,
        help="The location of the output file in which simulation data will be saved.",
    )
    action_arguments.add_argument(
        "--pv-system-size",
        "-pv",
        type=float,
        help="The size of the PV system being modelled in PV panel units, defaulting "
        "to kWp.",
    )
    action_arguments.add_argument(
        "--scenario",
        "-s",
        type=str,
        default=DEFAULT_SCENARIO,
        help="The name of the scenario to use for the run.",
    )
    action_arguments.add_argument(
        "--storage-size",
        "-b",
        type=float,
        help="The size of the battery system being modelled in battery units, "
        "defaulting to kWh.",
    )

    # Clean-water-specific arguments.
    clean_water_parser = parser.add_argument_group(
        "clean-water-only arguments",
    )
    clean_water_parser.add_argument(
        "--num-clean-water-tanks",
        "-ncwt",
        default=0,
        type=int,
        help="The number of clean-water tanks to be included in the system.",
    )
    clean_water_parser.add_argument(
        "--clean-water-pvt-system-size",
        "-cwpvt",
        type=float,
        help="The size of the PV-T system being modelled, associated with the "
        "clean-water supply system, in PV-T panel units.",
    )

    # Hot-water-specific arguments.
    hot_water_parser = parser.add_argument_group(
        "hot-water-only arguments",
    )
    hot_water_parser.add_argument(
        "--num-hot-water-tanks",
        "-nhwt",
        default=0,
        type=int,
        help="The number of hpt-water tanks to be included in the system.",
    )
    hot_water_parser.add_argument(
        "--hot-water-pvt-system-size",
        "-hwpvt",
        type=float,
        help="The size of the PV-T system being modelled, associated with the "
        "hot-water supply system, in PV-T panel units.",
    )

    # Simulation-specific arguments.
    simulation_parser = parser.add_argument_group(
        "simulation-only arguments",
    )
    simulation_parser.add_argument(
        "--simulation",
        "-sim",
        action="store_true",
        default=False,
        help="If specified, CLOVER will carry out a single simulation.",
    )
    simulation_parser.add_argument(
        "--analyse",
        "-a",
        action="store_true",
        default=False,
        help="If specified, plots will be generated and saved and key results will be "
        "calculated and saved for the simulation.",
    )
    simulation_parser.add_argument(
        "--skip-plots",
        "-sp",
        action="store_true",
        default=False,
        help="If specified, plots will be skipped. Note, only affects cases where "
        "analysis has been requested; plot generation cannot be carried out without "
        "analysis.",
    )

    # Optimisation arguments
    optimisation_parser = parser.add_argument_group(
        "optimisation-only arguments",
    )
    optimisation_parser.add_argument(
        "--optimisation",
        "-opt",
        action="store_true",
        default=False,
        help="If specified, CLOVER will carry out optimisations in accordance with "
        "the `optimisation_inputs` file.",
    )
    action_arguments.add_argument(
        "--optimisation-inputs-file",
        "-opt-file",
        type=str,
        help="The name of the optimisation inputs file to use for the run. This "
        "overrides CLOVER's in-built optimisation inputs filename.",
    )

    return parser.parse_args(args)


def validate_args(logger: logging.Logger, parsed_args: argparse.Namespace) -> bool:
    """
    Validates the command-line arguments passed in.

    Inputs:
        - parsed_args
            The parsed command-line arguments.

    Outputs:
        - A boolean giving whether the arguments are valid (True) or not (False).

    Raises: MissingParameterError
        - Raised when a CLI parameter is missing.

    """

    if parsed_args.location is None:
        logger.error(
            "%sThe required argument, 'location', was not specified.%s",
            BColours.fail,
            BColours.endc,
        )
        raise MissingParametersError("location")

    if parsed_args.simulation and parsed_args.optimisation:
        logger.error(
            "%sCannot run both a simulation and an optimisation.%s",
            BColours.fail,
            BColours.endc,
        )
        return False

    if parsed_args.electric_load_profile is not None and parsed_args.regenerate:
        logger.error(
            "%sCannot request profile regeneration if a total-load profile is "
            "specified.%s",
            BColours.fail,
            BColours.endc,
        )
        return False

    return True



#!/usr/bin/python3
########################################################################################
# storage_utils.py - Storage utility module.                                           #
#                                                                                      #
# Author: Ben Winchester                                                               #
# Copyright: Ben Winchester, 2022                                                      #
# Date created: 18/01/2022                                                             #
# License: Open source                                                                 #
# Most recent update: 18/01/2022                                                       #
########################################################################################
"""
storage_utils.py - The storage utility module for CLOVER.

CLOVER considers several storage media for various forms of energy. These are all
contained and considered within this module.

"""

import dataclasses

from typing import Any, Dict

from ..__utils__ import (
    HEAT_CAPACITY_OF_WATER,
    NAME,
    ResourceType,
)

__all__ = (
    "Battery",
    "CleanWaterTank",
    "HotWaterTank",
)


# Default storage unit:
#   The default unit size, in kWh, of the batteries being considered.
DEFAULT_STORAGE_UNIT = 1  # [kWh]


class _BaseStorage:
    """
    Repsesents an abstract base storage unit.

    .. attribute:: cycle_lifetime
        The number of cycles for which the :class:`_BaseStorage` can perform.

    .. attribute:: label
        The label given to the :class:`_BaseStorage` instance.

    .. attribute:: leakage
        The rate of level leakage from the :class:`_BaseStorage`.

    .. attribute:: maximum_charge
        The maximum level of the :class:`_BaseStorage`, defined between 0 (able to hold
        no charge) and 1 (able to fully charge).

    .. attribute:: minimum_charge
        The minimum level of the :class:`_BaseStorage`, defined between 0 (able to fully
        discharge) and 1 (unable to discharge any amount).

    .. attribute:: name
        A unique name for identifying the :class:`_BaseStorage`.

    .. attribute:: resource_type
        The type of resource being stored by the :class:`_BaseStorage` instance.

    """

    label: str
    resource_type: ResourceType

    def __init__(
        self,
        cycle_lifetime: int,
        leakage: float,
        maximum_charge: float,
        minimum_charge: float,
        name: str,
    ) -> None:
        """
        Instantiate a :class:`Storage` instance.

        Inputs:
            - cycle_lifetime:
                The number of cycles for which the :class:`Storage` instance can
                perform.
            - leakage:
                The rate of leakage from the storage.
            - maximum_charge:
                The maximum level that can be held by the :class:`_BaseStorage`.
            - minimum_charge:
                The minimum level to which the :class:`_BaseStorage` instance can
                discharge.
            - name:
                The name to assign to the :class:`Storage` instance.

        """

        self.cycle_lifetime: int = cycle_lifetime
        self.leakage: float = leakage
        self.maximum_charge: float = maximum_charge
        self.minimum_charge: float = minimum_charge
        self.name: str = name

    def __hash__(self) -> int:
        """
        Return a unique hash identifying the :class:`_BaseStorage` instance.

        Outputs:
            - Return a unique hash identifying the :class:`_BaseStorage` instance.

        """

        return hash(self.name)

    def __init_subclass__(cls, label: str, resource_type: ResourceType) -> None:
        """
        Method run when a :class:`_BaseStorage` child is instantiated.

        Inputs:
            - label:
                A `str` that identifies the class type.
            - resource_type:
                The type of load being modelled.

        """

        super().__init_subclass__()
        cls.label = label
        cls.resource_type = resource_type

    def __str__(self) -> str:
        """
        Returns a nice-looking string describing the :class:`_BaseStorage` instance.

        Outputs:
            - A `str` giving information about the :class:`_BaseStorage` instance.

        """

        return (
            "Storage("
            + f"name={self.name}, "
            + f"cycle_lifetime={self.cycle_lifetime} cycles, "
            + f"leakage={self.leakage}, "
            + f"maximum_charge={self.maximum_charge}, "
            + f"minimum_charge={self.minimum_charge}"
            + ")"
        )

    @classmethod
    def from_dict(cls, storage_data: Dict[str, Any]) -> Any:
        """
        Create a :class:`_BaseStorage` instance based on the file data passed in.

        Inputs:
            - storage_data:
                The storage data, extracted from the relevant input file.

        Outputs:
            - A :class:`_BaseStorage` instance.

        """

        return cls(
            storage_data["cycle_lifetime"],
            storage_data["leakage"],
            storage_data["maximum_charge"],
            storage_data["minimum_charge"],
            storage_data[NAME],
        )


@dataclasses.dataclass
class Battery(_BaseStorage, label="battery", resource_type=ResourceType.ELECTRIC):
    """
    Represents a battery within CLOVER.

    .. attribute:: capacity
        The capacity of the battery in kWh.

    .. attribute:: charge_rate
        The rate of charge of the :class:`Battery`.

    .. attribute:: conversion_in
        The input conversion efficiency of the :class:`Battery`.

    .. attribute:: conversion_out
        The output conversion efficiency of the :class:`Battery`.

    .. attribute:: discharge_rate
        The rate of discharge of the :class:`Battery`.

    .. attribute:: lifetime_loss
        The overall loss in capacity of the :class:`Battery` over its lifetime.

    .. attribute:: storage_unit
        The storage_unit of the :class:`Battery`, measured in kWh.

    """

    def __init__(
        self,
        capacity: float,
        cycle_lifetime: int,
        leakage: float,
        maximum_charge: float,
        minimum_charge: float,
        name: str,
        charge_rate: float,
        conversion_in: float,
        conversion_out: float,
        discharge_rate: float,
        lifetime_loss: float,
        storage_unit: float,
        storage_unit_overrided: bool,
    ) -> None:
        """
        Instantiate a :class:`Battery` instance.

        Inputs:
            - capacity:
                The capacity of the battery in kWh.
            - cycle_lifetime:
                The number of cycles for which the :class:`Battery` instance can
                perform.
            - leakage:
                The rate of leakage from the storage.
            - maximum_charge:
                The maximum level that can be held by the :class:`Battery`.
            - minimum_charge:
                The minimum level to which the :class:`Battery` instance can
                discharge.
            - name:
                The name to assign to the :class:`Battery` instance.
            - charge_rate:
                The rate of charge of the :class:`Battery`.
            - conversion_in:
                The efficiency of conversion of energy into the :class:`Battery`.
            - conversion_out:
                The efficiency of conversion of energy out of the :class:`Battery`.
            - discharge_rate:
                The rate of discharge of the :class:`Battery`.
            - lifetime_loss:
                The loss in capacity of the :class:`Battery` over its lifetime.
            - storage_unit:
                The storage_unit of the :class:`Battery` in kWh.
            - storage_unit_overrided:
                Whether the default storage unit has been overrided (True) or not
                (False).

        """

        super().__init__(cycle_lifetime, leakage, maximum_charge, minimum_charge, name)
        self.capacity: float = capacity
        self.charge_rate: float = charge_rate
        self.conversion_in: float = conversion_in
        self.conversion_out: float = conversion_out
        self.discharge_rate: float = discharge_rate
        self.lifetime_loss: float = lifetime_loss
        self.storage_unit = storage_unit
        self.storage_unit_overrided = storage_unit_overrided

    def __str__(self) -> str:
        """
        Returns a nice-looking string describing the :class:`_BaseStorage` instance.

        Outputs:
            - A `str` giving information about the :class:`_BaseStorage` instance.

        """

        return (
            "Battery("
            + f"{self.label} storing {self.resource_type.value} loads, "
            + f"name={self.name}, "
            + f"capacity={self.capacity}, "
            + f"cycle_lifetime={self.cycle_lifetime} cycles, "
            + f"leakage={self.leakage}, "
            + f"maximum_charge={self.maximum_charge}, "
            + f"minimum_charge={self.minimum_charge}, "
            + f"charge_rate={self.charge_rate}, "
            + f"discharge_rate={self.discharge_rate}, "
            + f"conversion_in={self.conversion_in}, "
            + f"conversion_out={self.conversion_out}, "
            + f"lifetime_loss={self.lifetime_loss}, "
            + f"size={self.storage_unit} kWh"
            + ")"
        )

    def __repr__(self) -> str:
        """
        Defines the default representation of the :class:`Battery` instance.

        Outputs:
            - A `str` giving the default representation of the :class:`Battery`
              instance.

        """

        return (
            "Battery("
            + f"{self.label} storing {self.resource_type.value} loads, "
            + f"name={self.name}, capacity={self.capacity}"
            + ")"
        )

    @classmethod
    def from_dict(cls, storage_data: Dict[str, Any]) -> Any:
        """
        Create a :class:`Battery` instance based on the file data passed in.

        Inputs:
            - storage_data:
                The battery data, extracted from the relevant input file.

        Outputs:
            - A :class:`Battery` instance.

        """

        return cls(
            storage_data["capacity"] if "capacity" in storage_data else 1,
            storage_data["cycle_lifetime"],
            storage_data["leakage"],
            storage_data["maximum_charge"],
            storage_data["minimum_charge"],
            storage_data[NAME],
            storage_data["c_rate_charging"],
            storage_data["conversion_in"],
            storage_data["conversion_out"],
            storage_data["c_rate_discharging"],
            storage_data["lifetime_loss"],
            storage_data["storage_unit"]
            if "storage_unit" in storage_data
            else DEFAULT_STORAGE_UNIT,
            "storage_unit" in storage_data,
        )


@dataclasses.dataclass
class CleanWaterTank(
    _BaseStorage, label="clean_water_tank", resource_type=ResourceType.CLEAN_WATER
):
    """
    Represents a clean-water tank within CLOVER.

    """

    def __init__(
        self,
        cycle_lifetime: int,
        leakage: float,
        maximum_charge: float,
        minimum_charge: float,
        name: str,
        mass: float,
    ) -> None:
        """
        Instantiate a :class:`CleanWaterTank`.

        Inputs:
            - cycle_lifetime:
                The number of cycles for which the :class:`CleanWaterTank` instance can
                perform.
            - leakage:
                The rate of leakage from the storage.
            - maximum_charge:
                The maximum level that can be held by the :class:`CleanWaterTank`.
            - minimum_charge:
                The minimum level to which the :class:`CleanWaterTank` instance can
                discharge.
            - name:
                The name to assign to the :class:`CleanWaterTank` instance.
            - mass:
                The mass of water that can be held in the clean-water tank.

        """

        super().__init__(cycle_lifetime, leakage, maximum_charge, minimum_charge, name)
        self.mass: float = mass

    def __str__(self) -> str:
        """
        Returns a nice-looking string describing the :class:`CleanWaterTank` instance.

        Outputs:
            - A `str` giving information about the :class:`CleanWaterTank` instance.

        """

        return (
            "CleanWaterTank("
            + f"{self.label} storing {self.resource_type.value} loads, "
            + f"name={self.name}, "
            + f"capacity={self.mass} litres, "
            + f"cycle_lifetime={self.cycle_lifetime} cycles, "
            + f"leakage={self.leakage}, "
            + f"maximum_charge={self.maximum_charge}, "
            + f"minimum_charge={self.minimum_charge}"
            + ")"
        )

    def __repr__(self) -> str:
        """
        Defines the default representation of the :class:`CleanWaterTank` instance.

        Outputs:
            - A `str` giving the default representation of the :class:`CleanWaterTank`
              instance.

        """

        return (
            "CleanWaterTank("
            + f"{self.label} storing {self.resource_type.value} loads, "
            + f"name={self.name}"
            + ")"
        )

    @classmethod
    def from_dict(cls, storage_data: Dict[str, Any]) -> Any:
        """
        Create a :class:`CleanWaterTank` instance based on the file data passed in.

        Inputs:
            - storage_data:
                The tank data, extracted from the relevant input file.

        Outputs:
            - A :class:`CleanWaterTank` instance.

        """

        return cls(
            storage_data["cycle_lifetime"],
            storage_data["leakage"],
            storage_data["maximum_charge"],
            storage_data["minimum_charge"],
            storage_data[NAME],
            storage_data["mass"],
        )


@dataclasses.dataclass
class HotWaterTank(
    CleanWaterTank, label="hot_water_tank", resource_type=ResourceType.HOT_CLEAN_WATER
):
    """
    Represents a hot-water tank within CLOVER.

    .. attribute:: area
        The area of the hot-water tank, measured in meters squared.

    .. attribute:: heat_capacity
        The specific heat capacity of the contents of the tank, measured in Joules per
        kilogram Kelvin, defaults to that of water at stp.

    .. attribute:: heat_loss_coefficient
        The heat loss from the tank, measured in Watts per meter squared per Kelvin.

    .. attribute:: heat_transfer_coefficient
        The heat transfer coefficient from the tank to its surroundings, measured in
        Watts per Kelvin.

    """

    def __init__(
        self,
        cycle_lifetime: int,
        leakage: float,
        maximum_charge: float,
        minimum_charge: float,
        name: str,
        mass: float,
        area: float,
        heat_capacity: float,
        heat_loss_coefficient: float,
    ) -> None:
        """
        Instantiate a :class:`CleanWaterTank`.

        Inputs:
            - cycle_lifetime:
                The number of cycles for which the :class:`HotWaterTank` instance can
                perform.
            - leakage:
                The rate of leakage from the storage.
            - maximum_charge:
                The maximum level that can be held by the :class:`HotWaterTank`.
            - minimum_charge:
                The minimum level to which the :class:`HotWaterTank` instance can
                discharge.
            - name:
                The name to assign to the :class:`HotWaterTank` instance.
            - mass:
                The mass of water that can be held in the clean-water tank.
            - area:
                The surface area of the tank.
            - heat_capacity:
                The specific heat capacity of the contents of the :class:`HotWaterTank`.
            - heta_loss_coefficient:
                The heat-loss coefficient for the :class:`HotWaterTank`.

        """

        super().__init__(
            cycle_lifetime, leakage, maximum_charge, minimum_charge, name, mass
        )
        self.area = area
        self.heat_capacity = heat_capacity
        self.heat_loss_coefficient = heat_loss_coefficient

    @property
    def heat_transfer_coefficient(self) -> float:
        """
        Return the heat-transfer coefficient from the :class:`HotWaterTank`.

        Outputs:
            - The heat-transfer coefficient from the :class:`HotWaterTank` to its
              surroundings, measured in Watts per Kelvin.

        """

        return self.heat_loss_coefficient * self.area

    def __str__(self) -> str:
        """
        Returns a nice-looking string describing the :class:`HotWaterTank` instance.

        Outputs:
            - A `str` giving information about the :class:`HotWaterTank` instance.

        """

        return (
            "HotWaterTank("
            + f"{self.label} storing {self.resource_type.value} loads, "
            + f"name={self.name}, "
            + f"area={self.area} m^2, "
            + f"capacity={self.mass} litres, "
            + f"cycle_lifetime={self.cycle_lifetime} cycles, "
            + f"heat_capacity={self.heat_capacity} J/kg*K, "
            + f"heat_loss_coefficient={self.heat_loss_coefficient} W/m^2K, "
            + f"heat_transfer_coefficient={self.heat_transfer_coefficient} W/K, "
            + f"leakage={self.leakage}, "
            + f"maximum_charge={self.maximum_charge}, "
            + f"minimum_charge={self.minimum_charge}"
            + ")"
        )

    def __repr__(self) -> str:
        """
        Defines the default representation of the :class:`HotWaterTank` instance.

        Outputs:
            - A `str` giving the default representation of the :class:`HotWaterTank`
              instance.

        """

        return (
            "HotWaterTank("
            + f"{self.label} storing {self.resource_type.value} loads, "
            + f"name={self.name}"
            + ")"
        )

    @classmethod
    def from_dict(cls, storage_data: Dict[str, Any]) -> Any:
        """
        Create a :class:`HotWaterTank` instance based on the file data passed in.

        Inputs:
            - storage_data:
                The tank data, extracted from the relevant input file.

        Outputs:
            - A :class:`HotWaterTank` instance.

        """

        return cls(
            storage_data["cycle_lifetime"],
            storage_data["leakage"],
            storage_data["maximum_charge"],
            storage_data["minimum_charge"],
            storage_data[NAME],
            storage_data["mass"],
            storage_data["area"],
            storage_data["heat_capacity"]
            if "heat_capacity" in storage_data
            else HEAT_CAPACITY_OF_WATER,
            storage_data["heat_loss_coefficient"],
        )



#!/usr/bin/env python
########################################################################################
# clover_hpc_outbox_assembly.py - Entry point for running CLOVER's HPC outbox assembly #
#                                                                                      #
# Author: Ben Winchester                                                               #
# Copyright: Ben Winchester, 2021                                                      #
# Date created: 08/03/2022                                                             #
# License: Open source                                                                 #
########################################################################################
"""
clover_hpc_outbox_assembly.py - Entry point for running CLOVER's HPC outbox script.

CLOVER can be run on Imperial College London's high-performance computers (HPC)s. In
order to do so as an installed piece of software, this wrapper provides an entry point
for the code once installed in order to assembly the optimisation output files.

"""

import sys

from .hpc_outbox_assembly import main as hpc_outbox_assembly_main


def main() -> None:
    """
    Main function of the CLOVER-HPC entry-point script.

    """

    hpc_outbox_assembly_main(sys.argv[1:])
